"""
This file containes functions to plot multiple scenarios in a folder
"""
import math
import os
import operator
import collections
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from energy_demand.plotting import plotting_styles
from energy_demand.plotting import basic_plot_functions
from energy_demand.basic import conversions
from energy_demand.plotting import plotting_results
from energy_demand import enduse_func
from energy_demand.profiles import load_factors
from energy_demand.read_write import write_data
from energy_demand.basic import date_prop
#matplotlib.use('Agg') # Used to make it work in linux

def plot_heat_pump_chart_multiple(
        lookups,
        regions,
        hp_scenario_data,
        fig_name,
        txt_name,
        fueltype_str_input,
        plotshow=False):
    """
    Compare share of element on x axis (provided in name of scenario)
    with load factor

    Info
    -----
    Run scenarios with different value in scenarion name

    e.g. 0.1 heat pump --> scen_0.1
    """
    # Year for which the whole is is generated
    year_to_plot = 2050

    # Collect value to display on axis
    result_dict = {} # {scenario_value:  {year: {fueltype: np.array(reg, value))}}
    results_txt = []

    for scenario_hp, scen_data in hp_scenario_data.items():
        result_dict[scenario_hp] = {}
        for scenario_name, scenario_data in scen_data.items():
            print("Scenario to process: " + str(scenario_name))

            # Scenario value
            value_scenario = int(scenario_name.split("__")[1])

            # Get peak for all regions {year: {fueltype: np.array(reg,value))}
            y_lf_fueltype = {}

            # Load factor
            '''for year, data_lf_fueltypes in scenario_data['reg_load_factor_y'].items(): # {scenario_value: np.array((regions, result_value))}
                if year != year_to_plot:
                    continue

                y_lf_fueltype[year] = {}
                for fueltype_int, data_lf in enumerate(data_lf_fueltypes):
                    fueltype_str = tech_related.get_fueltype_str(lookups['fueltypes'], fueltype_int)

                    # Select only fueltype data
                    if fueltype_str == fueltype_str_input:
                        y_lf_fueltype[year] = data_lf
                    else:
                        pass
            '''
            # PEAK VALUE ed_peak_regs_h ed_peak_h
            #for year, data_lf_fueltypes in scenario_data['ed_peak_regs_h'].items():
            for year, year_data in scenario_data['ed_peak_h'].items():
                if year != year_to_plot:
                    pass
                else:
                    y_lf_fueltype[year] = {}
                    for fueltype_str, peak_h in year_data.items():

                        # Select only fueltype data
                        if fueltype_str == fueltype_str_input:
                            y_lf_fueltype[year] = peak_h

                            txt = "fueltype_str: {}  peak_h: {} scenario_name: {} ".format(
                                fueltype_str,
                                peak_h,
                                scenario_name)

                            results_txt.append(txt)

                        else:
                            pass

            result_dict[scenario_hp][value_scenario] = y_lf_fueltype

        # Sort dict and convert to OrderedDict
        result_dict[scenario_hp] = collections.OrderedDict(sorted(result_dict[scenario_hp].items()))

    # --------------------------
    # Save model results to txt
    # --------------------------
    write_data.write_list_to_txt(
        txt_name,
        results_txt)

    #-----
    # Plot
    # -----
    plot_max_min_polygon = False
    plot_all_regs = False

    # Set figure size
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(9, 8))
    ax = fig.add_subplot(1, 1, 1)

    # -----------------
    # Axis
    # -----------------
    first_scenario = list(result_dict.keys())[0]

    # Percentages on x axis
    major_ticks = list(result_dict[first_scenario].keys())

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    color_list_selection = plotting_styles.get_colorbrewer_color(
        color_prop='sequential', #sequential
        color_palette='PuBu_4',
        inverse=False) # #https://jiffyclub.github.io/palettable/colorbrewer/sequential/

    # all percent values
    all_percent_values = list(result_dict[first_scenario].keys())

    # Nr of years
    for _percent_value, fuel_fueltype_yrs in result_dict[first_scenario].items():
        years = list(fuel_fueltype_yrs.keys())
        break

    for year in years:

        # ----------------
        # For every region
        # ----------------
        '''for reg_nr, _ in enumerate(regions):
            year_data = []
            for _percent_value, fuel_fueltype_yrs in result_dict.items():
                year_data.append(fuel_fueltype_yrs[year][reg_nr])

            # Paste out if not individual regions and set plot_max_min_polygon to True
            if plot_all_regs:
                plt.plot(
                    list(all_percent_values),
                    list(year_data),
                    color=str(color_scenario))'''

        # --------------------
        # Plot max min polygon
        # --------------------
        '''if plot_max_min_polygon:

            # Create value {x_vals: [y_vals]}
            x_y_values = {}
            for _percent_value, fuel_fueltype_yrs in result_dict.items():
                x_y_values[_percent_value] = []
                for reg_nr, _ in enumerate(regions):
                    x_y_values[_percent_value].append(result_dict[_percent_value][year])

            # Create polygons
            min_max_polygon = plotting_results.create_min_max_polygon_from_lines(x_y_values)

            polygon = plt.Polygon(
                min_max_polygon,
                color=color_scenario,
                alpha=0.2,
                edgecolor=None,
                linewidth=0,
                fill='True')

            ax.add_patch(polygon)'''
        color_scenarios = plotting_styles.color_list_scenarios()

        cnt = -1
        for scenario_hp, scenario_hp_result in result_dict.items():
            print("Scneario: " + str(scenario_hp))
            cnt += 1
            # Average across all regs
            year_data = []
            for _percent_value, fuel_fueltype_yrs in scenario_hp_result.items():

                regs = fuel_fueltype_yrs[year]

                # --------------------------------------
                # Average load factor across all regions
                # --------------------------------------
                lf_peak_across_all_regs = np.average(regs)
                year_data.append(lf_peak_across_all_regs)

            plt.plot(
                list(all_percent_values),
                list(year_data),
                label=scenario_hp,
                color=color_scenarios[cnt])
    

    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)
    plt.ylim(ymax=80)

    plt.xlim(xmin=0)
    plt.xlim(xmax=100)

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        #legend_entries,
        ncol=1,
        loc=3,
        prop={
            'family': 'arial',
            'size': 5},
        frameon=False)

    # ---------
    # Labels
    # ---------
    plt.xlabel("heat pump residential heating [%]")
    #plt.ylabel("load factor [%] [{}]".format(fueltype_str_input))
    plt.ylabel("Peak demand h [GW] {}".format(fueltype_str_input))
    #plt.title("impact of changing residential heat pumps to load factor")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    plt.close()

def plot_heat_pump_chart(
        lookups,
        regions,
        scenario_data,
        fig_name,
        fueltype_str_input,
        plotshow=False):
    """
    Compare share of element on x axis (provided in name of scenario)
    with load factor


    Info
    -----
    Run scenarios with different value in scenarion name

    e.g. 0.1 heat pump --> scen_0.1
    """
    year_to_plot = 2050

    # Collect value to display on axis
    result_dict = {} # {scenario_value:  {year: {fueltype: np.array(reg, value))}}

    for scenario_name, scenario_data in scenario_data.items():
        print("Scenario to process: " + str(scenario_name))
        # Scenario value
        value_scenario = float(scenario_name.split("__")[1])

        # Get peak for all regions {year: {fueltype: np.array(reg,value))}
        y_lf_fueltype = {}

        # Load factor
        '''for year, data_lf_fueltypes in scenario_data['reg_load_factor_y'].items(): # {scenario_value: np.array((regions, result_value))}
            if year != year_to_plot:
                continue

            y_lf_fueltype[year] = {}
            for fueltype_int, data_lf in enumerate(data_lf_fueltypes):
                fueltype_str = tech_related.get_fueltype_str(lookups['fueltypes'], fueltype_int)

                # Select only fueltype data
                if fueltype_str == fueltype_str_input:
                    y_lf_fueltype[year] = data_lf
                else:
                    pass
        '''
        # PEAK VALUE ed_peak_regs_h ed_peak_h
        #for year, data_lf_fueltypes in scenario_data['ed_peak_regs_h'].items():
        for year, data_lf_fueltypes in scenario_data['ed_peak_h'].items():
            if year != year_to_plot:
                continue

            y_lf_fueltype[year] = {}
            for fueltype_str, data_lf in data_lf_fueltypes.items():

                # Select only fueltype data
                if fueltype_str == fueltype_str_input:
                    y_lf_fueltype[year] = data_lf
                else:
                    pass

        result_dict[value_scenario] = y_lf_fueltype

    # Sort dict and convert to OrderedDict
    result_dict = collections.OrderedDict(sorted(result_dict.items()))

    #-----
    # Plot
    # -----

    # Criteria to plot maximum boundaries
    plot_max_min_polygon = False
    plot_all_regs = False

    # Set figure size
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(16, 8))
    ax = fig.add_subplot(1, 1, 1)

    # -----------------
    # Axis
    # -----------------
    # Percentages on x axis
    major_ticks = list(result_dict.keys())

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    color_list_selection = plotting_styles.get_colorbrewer_color(
        color_prop='sequential', #sequential
        color_palette='PuBu_4',
        inverse=False) # #https://jiffyclub.github.io/palettable/colorbrewer/sequential/

    # all percent values
    all_percent_values = list(result_dict.keys())

    # Nr of years
    for _percent_value, fuel_fueltype_yrs in result_dict.items():
        years = list(fuel_fueltype_yrs.keys())
        break

    legend_entries = []
    for year in years:
        color_scenario = color_list_selection.pop()

        legend_entries.append("mean {}".format(year))

        # ----------------
        # For every region
        # ----------------
        '''for reg_nr, _ in enumerate(regions):
            year_data = []
            for _percent_value, fuel_fueltype_yrs in result_dict.items():
                year_data.append(fuel_fueltype_yrs[year][reg_nr])

            # Paste out if not individual regions and set plot_max_min_polygon to True
            if plot_all_regs:
                plt.plot(
                    list(all_percent_values),
                    list(year_data),
                    color=str(color_scenario))'''

        # --------------------
        # Plot max min polygon
        # --------------------
        if plot_max_min_polygon:

            # Create value {x_vals: [y_vals]}
            x_y_values = {}
            for _percent_value, fuel_fueltype_yrs in result_dict.items():
                x_y_values[_percent_value] = []
                for reg_nr, _ in enumerate(regions):
                    x_y_values[_percent_value].append(result_dict[_percent_value][year])

            # Create polygons
            min_max_polygon = plotting_results.create_min_max_polygon_from_lines(x_y_values)

            polygon = plt.Polygon(
                min_max_polygon,
                color=color_scenario,
                alpha=0.2,
                edgecolor=None,
                linewidth=0,
                fill='True')

            ax.add_patch(polygon)

        # Average across all regs
        year_data = []
        for _percent_value, fuel_fueltype_yrs in result_dict.items():

            regs = fuel_fueltype_yrs[year]

            # --------------------------------------
            # Average load factor across all regions
            # --------------------------------------
            lf_peak_across_all_regs = np.average(regs)
            year_data.append(lf_peak_across_all_regs)

        plt.plot(
            list(all_percent_values),
            list(year_data),
            color=str(color_scenario))
    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)
    plt.ylim(ymax=80)
    #plt.ylim(ymax=1.2)
    plt.xlim(xmin=0, xmax=60)

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        legend_entries,
        ncol=1,
        loc=3,
        prop={
            'family': 'arial',
            'size': 10},
        frameon=False)

    # ---------
    # Labels
    # ---------
    plt.xlabel("heat pump residential heating [%]")
    #plt.ylabel("load factor [%] [{}]".format(fueltype_str_input))
    plt.ylabel("Peak demand h [GW] {}".format(fueltype_str_input))
    #plt.title("impact of changing residential heat pumps to load factor")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()

def plot_tot_y_peak_hour(
        scenario_data,
        fig_name,
        fueltype_str_input,
        plotshow=False
    ):
    """Plot fueltype specific peak h of all regions
    """
    plt.figure(figsize=basic_plot_functions.cm2inch(14, 8))

    # -----------------
    # Axis
    # -----------------
    base_yr, year_interval = 2015, 5
    first_scen = list(scenario_data.keys())[0]
    end_yr = list(scenario_data[first_scen]['ed_peak_h'].keys())[-1]

    major_ticks = np.arange(
        base_yr,
        end_yr + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    color_list_selection = plotting_styles.color_list_selection()

    for scenario_name, fuel_fueltype_yrs in scenario_data.items():

        data_container = []
        for year, fuel_fueltypes in fuel_fueltype_yrs['ed_peak_h'].items():
            data_container.append(fuel_fueltypes[fueltype_str_input])

        # Calculate max peak in end year
        peak_gw = round(data_container[-1], 1)
        peak_p = round((100 / data_container[0]) * data_container[-1], 1)

        # Label
        label_scenario = "{} (peak: GW {} p: {})".format(
            scenario_name,
            peak_gw,
            peak_p)

        plt.plot(
            list(fuel_fueltype_yrs['ed_peak_h'].keys()),     # years
            list(data_container),               # yearly data
            color=str(color_list_selection.pop()),
            label=label_scenario) #scenario_name)

    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        ncol=1,
        loc=3,
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)

    # ---------
    # Labels
    # ---------
    plt.ylabel("GWh")
    plt.xlabel("year")
    plt.title("peak_h {} [GW]".format(fueltype_str_input))

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()

def plot_reg_y_over_time(
        scenario_data,
        fig_name,
        plotshow=False
    ):
    """Plot total demand over simulation period for every
    scenario for all regions
    """
    # Set figure size
    plt.figure(figsize=basic_plot_functions.cm2inch(14, 8))

    y_scenario = {}

    for scenario_name, scen_data in scenario_data.items():

        data_years_regs = {}
        for year, fueltype_reg_time in scen_data['ed_fueltype_regs_yh'].items():
            data_years_regs[year] = {}

            for _fueltype, regions_fuel in enumerate(fueltype_reg_time):
                for region_nr, region_fuel in enumerate(regions_fuel):

                    # Sum all regions and fueltypes
                    reg_gwh_fueltype_y = np.sum(region_fuel)

                    try:
                        data_years_regs[year][region_nr] += reg_gwh_fueltype_y
                    except:
                        data_years_regs[year][region_nr] = reg_gwh_fueltype_y

        y_scenario[scenario_name] = data_years_regs

    # -----------------
    # Axis
    # -----------------
    base_yr, year_interval = 2015, 5
    first_scen = list(y_scenario.keys())[0]
    end_yr = list(y_scenario[first_scen].keys())[-1]

    major_ticks = np.arange(
        base_yr,
        end_yr + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    color_list_selection = plotting_styles.color_list_scenarios()
    cnt = -1
    for scenario_name, fuel_fueltype_yrs in y_scenario.items():

        cnt += 1

        color_scenario = color_list_selection[cnt]

        for year, regs in fuel_fueltype_yrs.items():
            nr_of_reg = len(regs.keys())
            break

        for reg_nr in range(nr_of_reg):
            reg_data = []
            for year, regions_fuel in fuel_fueltype_yrs.items():
                reg_data.append(regions_fuel[reg_nr])

            plt.plot(
                list(fuel_fueltype_yrs.keys()),
                list(reg_data),
                label="{}".format(scenario_name),
                color=str(color_scenario))
    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)

    # ------------
    # Plot legend
    # ------------
    '''plt.legend(
        ncol=2,
        loc=3,
        prop={
            'family': 'arial',
            'size': 10},
        frameon=False)'''

    # ---------
    # Labels
    # ---------
    font_additional_info = plotting_styles.font_info(size=5)

    plt.ylabel("GWh")
    plt.xlabel("year")
    plt.title(
        "tot_y",
        fontdict=font_additional_info)

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()

def plot_tot_fueltype_y_over_time(
        scenario_data,
        fueltypes,
        fueltypes_to_plot,
        fig_name,
        txt_name,
        plotshow=False
    ):
    """Plot total demand over simulation period for every
    scenario for all regions
    """
    results_txt = []

    # Set figure size
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(9, 8))

    ax = fig.add_subplot(1, 1, 1)

    y_scenario = {}

    for scenario_name, scen_data in scenario_data.items():

        # Read out fueltype specific max h load
        data_years = {}
        for year, fueltype_reg_time in scen_data['ed_fueltype_regs_yh'].items():

            # Sum all regions
            tot_gwh_fueltype_yh = np.sum(fueltype_reg_time, axis=1)

            # Sum all hours
            tot_gwh_fueltype_y = np.sum(tot_gwh_fueltype_yh, axis=1)

            # Convert to TWh
            tot_gwh_fueltype_y = conversions.gwh_to_twh(tot_gwh_fueltype_y)

            data_years[year] = tot_gwh_fueltype_y

        y_scenario[scenario_name] = data_years

    # -----------------
    # Axis
    # -----------------
    base_yr, year_interval = 2020, 10
    first_scen = list(y_scenario.keys())[0]
    end_yr = list(y_scenario[first_scen].keys())

    major_ticks = np.arange(
        base_yr,
        end_yr[-1] + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    #color_list_selection_fueltypes = plotting_styles.color_list_selection()
    color_list_selection = plotting_styles.color_list_scenarios()
    linestyles = ['--', '-', ':', "-.", ".-"] #linestyles = plotting_styles.linestyles()

    cnt_scenario = -1
    for scenario_name, fuel_fueltype_yrs in y_scenario.items():
        cnt_scenario += 1
        color = color_list_selection[cnt_scenario]
        cnt_linestyle = -1

        for fueltype_str, fueltype_nr in fueltypes.items():

            if fueltype_str in fueltypes_to_plot:
                cnt_linestyle += 1
                # Get fuel per fueltpye for every year
                fuel_fueltype = []
                for entry in list(fuel_fueltype_yrs.values()):
                    fuel_fueltype.append(entry[fueltype_nr])

                plt.plot(
                    list(fuel_fueltype_yrs.keys()),      # years
                    fuel_fueltype,                       # yearly data per fueltype
                    color=color,
                    linestyle=linestyles[cnt_linestyle],
                    label="{}_{}".format(scenario_name, fueltype_str))

                # ---
                # Calculate difference in demand from 2015 - 2050
                # ---
                tot_2015 = fuel_fueltype[0]
                tot_2050 = fuel_fueltype[-1]
                diff_abs = tot_2050 - tot_2015
                p_diff_2015_2015 = ((100 / tot_2015) * tot_2050) - 100 # Diff
                
                txt = "fueltype_str: {} ed_tot_by: {} ed_tot_cy: {}, diff_p: {}, diff_abs: {} scenario: {}".format(
                    fueltype_str,
                    round(tot_2015, 1),
                    round(tot_2050, 1),
                    round(p_diff_2015_2015, 1),
                    round(diff_abs, 1),
                    scenario_name)

                results_txt.append(txt)

    # --------------------------
    # Save model results to txt
    # --------------------------
    write_data.write_list_to_txt(
        txt_name,
        results_txt)

    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)

    # ------------
    # Plot legend
    # ------------
    ax.legend(
        ncol=2,
        frameon=False,
        loc='upper center',
        prop={
            'family': 'arial',
            'size': 4},
        bbox_to_anchor=(0.5, -0.1))

    # ---------
    # Labels
    # ---------
    font_additional_info = plotting_styles.font_info(size=5)

    plt.ylabel("TWh")
    plt.xlabel("year")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()

    plt.close()

def plot_tot_y_over_time(
        scenario_data,
        fig_name,
        plotshow=False
    ):
    """Plot total demand over simulation period for every
    scenario for all regions
    """
    # Set figure size
    plt.figure(figsize=basic_plot_functions.cm2inch(14, 8))

    y_scenario = {}

    for scenario_name, scen_data in scenario_data.items():

        # Read out fueltype specific max h load
        data_years = {}
        for year, fueltype_reg_time in scen_data['ed_fueltype_regs_yh'].items():

            # Sum all regions and fueltypes
            tot_gwh_fueltype_y = np.sum(fueltype_reg_time)

            # Convert to TWh
            tot_twh_fueltype_y = conversions.gwh_to_twh(tot_gwh_fueltype_y)

            data_years[year] = tot_twh_fueltype_y

        y_scenario[scenario_name] = data_years

    # -----------------
    # Axis
    # -----------------
    base_yr, year_interval = 2015, 5
    first_scen = list(y_scenario.keys())[0]
    end_yr = list(y_scenario[first_scen].keys())

    major_ticks = np.arange(
        base_yr,
        end_yr[-1] + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
   # color_list_selection = plotting_styles.color_list_selection()
    color_list_selection = plotting_styles.color_list_scenarios()

    for scenario_name, fuel_fueltype_yrs in y_scenario.items():

        plt.plot(
            list(fuel_fueltype_yrs.keys()),     # years
            list(fuel_fueltype_yrs.values()),   # yearly data per fueltype
            color=str(color_list_selection.pop()),
            label=scenario_name)

    # ----
    # Axis
    # ----
    plt.ylim(ymin=0)

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        ncol=1,
        loc=3,
        prop={
            'family': 'arial',
            'size': 10},
        frameon=False)

    # ---------
    # Labels
    # ---------
    plt.ylabel("TWh")
    plt.xlabel("year")
    plt.title("tot y ED all fueltypes")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()

    plt.close()

def plot_radar_plots_average_peak_day(
        scenario_data,
        fueltype_to_model,
        fueltypes,
        year_to_plot,
        fig_name,
        base_yr=2015
    ):
    """Compare averaged dh profile overall regions for peak day
    for future year and base year

    MAYBE: SO FAR ONLY FOR ONE SCENARIO
    """
    fueltype_int = fueltypes[fueltype_to_model]

    # ----------------
    # Create base year peak load profile
    # Aggregate load profiles of all regions
    # -----------------
    peak_winter_individ_radars_to_plot_dh = []
    trough_summer_individ_radars_to_plot_dh = []
    load_factor_fueltype_y_cy = []
    results_txt = []

    for scenario_cnt, scenario in enumerate(scenario_data):

        print("-------Scenario: '{}' '{}'".format(scenario, fueltype_to_model))

        # ------------------------
        # Future year load profile
        # ------------------------
        all_regs_fueltypes_yh_by = np.sum(scenario_data[scenario]['ed_fueltype_regs_yh'][base_yr], axis=1)
        all_regs_fueltypes_yh_cy = np.sum(scenario_data[scenario]['ed_fueltype_regs_yh'][year_to_plot], axis=1)

        # ---------------------------
        # Calculate load factors
        # ---------------------------
        peak_day_nr_by, by_max_h = enduse_func.get_peak_day_single_fueltype(
            all_regs_fueltypes_yh_by[fueltype_int])
        test_peak_day_by = enduse_func.get_peak_day_all_fueltypes(all_regs_fueltypes_yh_by)
        peak_day_nr_cy, cy_max_h = enduse_func.get_peak_day_single_fueltype(
            all_regs_fueltypes_yh_cy[fueltype_int])
        test_peak_day_cy = enduse_func.get_peak_day_all_fueltypes(all_regs_fueltypes_yh_cy)

        #peak_day_nr_cy = 32

        _ = all_regs_fueltypes_yh_cy[fueltype_int].reshape(365,24)
        print(_[peak_day_nr_cy])
        print(np.max(_[peak_day_nr_cy]))
        print(cy_max_h)

        print("-----{}  {}".format(scenario, peak_day_nr_cy))
        print(_[peak_day_nr_cy])
        print("EGON")

        # Minimum trough day
        trough_day_nr_by, by_max_h_trough = enduse_func.get_trough_day_single_fueltype(
            all_regs_fueltypes_yh_by[fueltype_int])

        trough_day_nr_cy, cy_max_h_trough = enduse_func.get_trough_day_single_fueltype(
            all_regs_fueltypes_yh_cy[fueltype_int])

        scen_load_factor_fueltype_y_by = load_factors.calc_lf_y(all_regs_fueltypes_yh_by)
        load_factor_fueltype_y_by = round(scen_load_factor_fueltype_y_by[fueltype_int], fueltype_int)

        scen_load_factor_fueltype_y_cy = load_factors.calc_lf_y(all_regs_fueltypes_yh_cy)
        load_factor_fueltype_y_cy.append(round(scen_load_factor_fueltype_y_cy[fueltype_int], fueltype_int))

        # Add info on peak days #by2: {} 
        results_txt.append("p_day: {} by1: {} cy1: {}: cy2: {} trough_by: {} trough_cy: {}".format(
            peak_day_nr_cy,
            date_prop.yearday_to_date(2015, peak_day_nr_by), 
            date_prop.yearday_to_date(2015, peak_day_nr_cy),
            #date_prop.yearday_to_date(2015, test_peak_day_by),
            date_prop.yearday_to_date(2015, test_peak_day_cy),
            date_prop.yearday_to_date(2015, trough_day_nr_by),
            date_prop.yearday_to_date(2015, trough_day_nr_cy)))

        # ------------------------
        # Restult calculations
        # ------------------------
        # Calculate share or space and water heating of total electrictiy demand in 2050
        enduses_to_agg = [
            'rs_space_heating',
            'rs_water_heating',
            'ss_space_heating',
            'ss_water_heating',
            'is_space_heating']

        aggregated_enduse_fueltype_cy = np.zeros((8760))
        aggregated_enduse_fueltype_by = np.zeros((8760))

        for enduse in scenario_data[scenario]['results_enduse_every_year'][year_to_plot].keys():
            if enduse in enduses_to_agg:
                aggregated_enduse_fueltype_by += scenario_data[scenario]['results_enduse_every_year'][2015][enduse][fueltype_int]
                aggregated_enduse_fueltype_cy += scenario_data[scenario]['results_enduse_every_year'][year_to_plot][enduse][fueltype_int]
        aggregated_enduse_fueltype_by_8760 = aggregated_enduse_fueltype_by.reshape(365, 24)
        aggregated_enduse_fueltype_cy_8760 = aggregated_enduse_fueltype_cy.reshape(365, 24)

        # Total demand of selected enduses
        selected_enduses_peak_by = round(np.max(aggregated_enduse_fueltype_by_8760[peak_day_nr_by]), 1)
        selected_enduses_peak_cy = round(np.max(aggregated_enduse_fueltype_cy_8760[peak_day_nr_cy]), 1)
        selected_enduses_min_by = round(np.max(aggregated_enduse_fueltype_by_8760[trough_day_nr_by]), 1)
        selected_enduses_min_cy = round(np.max(aggregated_enduse_fueltype_cy_8760[trough_day_nr_cy]), 1)

        # Calculate change in peak
        all_regs_fueltypes_yh_by = all_regs_fueltypes_yh_by.reshape(all_regs_fueltypes_yh_by.shape[0], 365, 24)
        all_regs_fueltypes_yh_cy = all_regs_fueltypes_yh_cy.reshape(all_regs_fueltypes_yh_cy.shape[0], 365, 24)

        diff_max_h = round(((100 / by_max_h) * cy_max_h) - 100, 1)

        txt = "peak_h_by: {} peak_h_cy: {} winter_heating_by: {} winter_heating_cy: {} trough_by: {} trough_cy: {} summer_heating_by: {} summer_heating_cy: {} {}".format(
            round(by_max_h, 1),
            round(cy_max_h, 1),
            selected_enduses_peak_by,
            selected_enduses_peak_cy,
            round(by_max_h_trough, 1),
            round(cy_max_h_trough, 1),
            selected_enduses_min_by,
            selected_enduses_min_cy,
            scenario)
    
        results_txt.append(txt)
        results_txt.append("----------")
        #results_txt.append("trough:day: " + str(all_regs_fueltypes_yh_by[fueltype_int][trough_day_nr_by]))

        print("Calculation of diff in peak: {} {} {} {}".format(
            scenario, round(diff_max_h, 1), round(by_max_h, 1), round(cy_max_h, 1)))

        # ----------------------------------
        # Plot dh for peak day for base year
        # ----------------------------------
        if scenario_cnt == 0:
            peak_winter_individ_radars_to_plot_dh.append(list(all_regs_fueltypes_yh_by[fueltype_int][peak_day_nr_by]))
            trough_summer_individ_radars_to_plot_dh.append(list(all_regs_fueltypes_yh_by[fueltype_int][trough_day_nr_cy]))
        else:
            pass

        # Add current year maximum winter peak day
        peak_winter_individ_radars_to_plot_dh.append(list(all_regs_fueltypes_yh_cy[fueltype_int][peak_day_nr_cy]))

        # Add current year minimum summer trough day
        trough_summer_individ_radars_to_plot_dh.append(list(all_regs_fueltypes_yh_cy[fueltype_int][trough_day_nr_cy]))
    
    # --------------------------
    # Save model results to txt
    # --------------------------
    name_spider_plot_peak = os.path.join(
        fig_name, "spider_scenarios_{}.pdf".format(fueltype_to_model))

    write_data.write_list_to_txt(
        os.path.join(fig_name, name_spider_plot_peak[:-3] + "txt"),
        results_txt)

    name_spider_plot_trough = os.path.join(
        fig_name, "spider_scenarios_min_summer{}.pdf".format(fueltype_to_model))

    write_data.write_list_to_txt(
        os.path.join(fig_name, name_spider_plot_trough[:-3] + "txt"),
        results_txt)

    # --------------------------
    # Plot figure
    # --------------------------

    # PEAK WINTER DAY PLOTS
    plot_radar_plot_multiple_lines(
        peak_winter_individ_radars_to_plot_dh,
        name_spider_plot_peak,
        plot_steps=50,
        scenario_names=list(scenario_data.keys()),
        plotshow=False,
        lf_y_by=[],
        lf_y_cy=[],
        list_diff_max_h=results_txt)

    # TROUGH Summer DAY PLOTS
    plot_radar_plot_multiple_lines(
        trough_summer_individ_radars_to_plot_dh,
        name_spider_plot_trough,
        plot_steps=50,
        scenario_names=list(scenario_data.keys()),
        plotshow=False,
        lf_y_by=[],
        lf_y_cy=[],
        list_diff_max_h=results_txt)

def plot_LAD_comparison_scenarios(
        scenario_data,
        year_to_plot,
        fig_name,
        plotshow=True
    ):
    """Plot chart comparing total annual demand for all LADs

    Arguments
    ---------
    scenario_data : dict
        Scenario name, scenario data
    year_to_plot : int
        Year to plot different LAD values
    fig_name : str
        Path to out pdf figure
    plotshow : bool
        Plot figure or not

    Info
    -----
    if scenario name starts with _ the legend does not work
    """

    # Get first scenario in dict
    all_scenarios = list(scenario_data.keys())
    first_scenario = str(all_scenarios[:1][0])

    # ----------------
    # Sort regions according to size
    # -----------------
    regions = {}
    for fueltype, fuels_regs in enumerate(scenario_data[first_scenario]['ed_fueltype_regs_yh'][2015]):

        for region_array_nr, fuel_reg in enumerate(fuels_regs):
            try:
                regions[region_array_nr] += np.sum(fuel_reg)
            except KeyError:
                regions[region_array_nr] = np.sum(fuel_reg)

    sorted_regions = sorted(
        regions.items(),
        key=operator.itemgetter(1))

    sorted_regions_nrs = []
    for sort_info in sorted_regions:
        sorted_regions_nrs.append(sort_info[0])

    # Labels
    labels = []
    for sorted_region in sorted_regions_nrs:
        geocode_lad = sorted_region # If actual LAD name, change this
        labels.append(geocode_lad)

    # -------------------------------------
    # Plot
    # -------------------------------------
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(9, 8))

    ax = fig.add_subplot(1, 1, 1)

    x_values = np.arange(0, len(sorted_regions_nrs), 1)

    # ----------------------------------------------
    # Plot base year values
    # ----------------------------------------------
    base_year_data = []
    for reg_array_nr in sorted_regions_nrs:
        base_year_data.append(regions[reg_array_nr])
    total_base_year_sum = sum(base_year_data)

    plt.plot(
        x_values,
        base_year_data,
        linestyle='None',
        marker='o',
        markersize=1.6,
        fillstyle='full',
        markerfacecolor='grey',
        markeredgewidth=0.4,
        color='black',
        label='actual_by ({})'.format(total_base_year_sum))

    # ----------------------------------------------
    # Plot all future scenario values
    # ----------------------------------------------
    color_list = plotting_styles.color_list()

    for scenario_nr, (scenario_name, fuel_data) in enumerate(scenario_data.items()):

        sorted_year_data = []
        for reg_array_nr in sorted_regions_nrs:
            tot_fuel_across_fueltypes = 0
            for fueltype, fuel_regs in enumerate(fuel_data['ed_fueltype_regs_yh'][year_to_plot]):
                tot_fuel_across_fueltypes += np.sum(fuel_regs[reg_array_nr])

            sorted_year_data.append(tot_fuel_across_fueltypes)

        tot_fuel_all_reg = np.sum(fuel_data['ed_fueltype_regs_yh'][year_to_plot])
        print("TOTAL FUEL in GWH " + str(tot_fuel_all_reg))

        # Calculate total annual demand
        tot_demand = sum(sorted_year_data)
        scenario_name = "{} (tot: {} [GWh])".format(
            scenario_name, round(tot_demand, 2))

        plt.plot(
            x_values,
            sorted_year_data,
            linestyle='None',
            marker='o',
            markersize=1.6,
            fillstyle='full',
            markerfacecolor=color_list[scenario_nr],
            markeredgewidth=0.4,
            color=color_list[scenario_nr],
            label=scenario_name)

    # --------
    # Axis
    # --------
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='off') # labels along the bottom edge are off

    # Limit
    plt.ylim(ymin=0)

    # -----------
    # Labelling
    # -----------
    label_points = False
    if label_points:
        for pos, txt in enumerate(labels):
            ax.text(
                x_values[pos],
                sorted_year_data[pos],
                txt,
                horizontalalignment="right",
                verticalalignment="top",
                fontsize=3)

    plt.title(
        "TEST",
        loc='left',
        fontdict=plotting_styles.font_info())

    plt.xlabel("UK regions (excluding northern ireland)")
    plt.ylabel("[GWh]")

    # --------
    # Legend
    # --------
    plt.legend(
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()
    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    else:
        plt.close()

def plot_radar_plot_multiple_lines(
        dh_profiles,
        fig_name,
        plot_steps,
        scenario_names,
        plotshow=False,
        lf_y_by=None,
        lf_y_cy=None,
        list_diff_max_h=None
    ):
    """Plot daily load profile on a radar plot

    Arguments
    ---------
    dh_profile : list
        Dh values to plot
    fig_name : str
        Path to save figure

    SOURCE: https://python-graph-gallery.com/390-basic-radar-chart/
    """
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(9, 14))

    # Get maximum demand of all lines
    max_entry = 0
    for line_entries in dh_profiles:
        max_in_line = max(line_entries)
        if max_in_line > max_entry:
            max_entry = max_in_line

    max_demand = round(max_entry, -1) + 10 # Round to nearest 10 plus add 10

    nr_of_plot_steps = int(max_demand / plot_steps) + 1

    axis_plots_inner = []
    axis_plots_innter_position = []

    # --------------------
    # Ciruclar axis
    # --------------------
    for i in range(nr_of_plot_steps):
        axis_plots_inner.append(plot_steps*i)
        axis_plots_innter_position.append(str(plot_steps*i))

    # Minor ticks
    minor_tick_interval = plot_steps / 2
    minor_ticks = []
    for i in range(nr_of_plot_steps * 2):
        minor_ticks.append(minor_tick_interval * i)

    # ----
    # Select color scheme
    # ----

    # Colors with scenarios
    #color_scenarios = plotting_styles.color_list_scenarios() #Color scheme Fig 13

    # Colors for plotting Fig. 13
    #color_scenarios = plotting_styles.color_list_selection_dm() # Color scheme Fig 13
    
    # Color sheme multiple scenarios
    color_scenarios = ['orange', 'darkred',  '#3b2c85', '#85cfcb'] #Lines for Fig 12

    color_lines = ['black'] + color_scenarios
    #color_lines = color_scenarios
    years = ['2015', '2050']
    linewidth_list = [1.0, 0.7]
    linestyle_list = ['-', '--']
    scenario_names.insert(0, "any")

    # Iterate lines
    cnt = -1
    for dh_profile in dh_profiles:

        if cnt >= 0:
            cnt_year = 1
        else:
            cnt_year = 0
        cnt += 1

        # Line properties
        color_line = color_lines[cnt]
        year_line = years[cnt_year]

        data = {'dh_profile': ['testname']}

        # Key: Label outer circle
        for hour in range(24):
            data[hour] = dh_profile[hour]

        # Set data
        df = pd.DataFrame(data)

        # number of variable
        categories = list(df)[1:]
        N = len(categories)

        # We are going to plot the first line of the data frame.
        # But we need to repeat the first value to close the circular graph:
        values = df.loc[0].drop('dh_profile').values.flatten().tolist()
        values += values[:1]

        # What will be the angle of each axis in the plot? (we divide the plot / number of variable)
        angles = [n / float(N) * 2 * math.pi for n in range(N)]
        angles += angles[:1]

        # Initialise the spider plot
        ax = plt.subplot(111, polar=True)

        # Change axis properties
        ax.yaxis.grid(color='grey', linestyle=':', linewidth=0.4)
        ax.xaxis.grid(color='lightgrey', linestyle=':', linewidth=0.4)

        # Change to clockwise cirection
        ax.set_theta_direction(-1)

        # Set first hour on top
        ax.set_theta_zero_location("N")

        # Draw one axe per variable + add labels labels yet (numbers)
        plt.xticks(
            angles[:-1],
            categories,
            color='black',
            size=8)

        # Draw ylabels (numbers)
        ax.set_rlabel_position(0)

        # Working alternative
        plt.yticks(
            axis_plots_inner,
            axis_plots_innter_position,
            color="black",
            size=8)

        #ax.set_yticks(axis_plots_inner, minor=False)
        #ax.tick_params(axis='y', pad=35) 
        #ax.set_yticks(minor_ticks, minor=True) #Somehow not working

        # Set limit to size
        plt.ylim(0, max_demand)
        plt.ylim(0, nr_of_plot_steps*plot_steps)

        # Smooth lines
        angles_smoothed, values_smoothed = basic_plot_functions.smooth_data(
            angles, values, spider=True)

        # Plot data
        ax.plot(
            angles_smoothed,
            values_smoothed,
            color=color_line,
            linestyle=linestyle_list[cnt_year],
            linewidth=linewidth_list[cnt_year],
            label="{} {}".format(year_line, scenario_names[cnt]))

        # Radar area
        ax.fill(
            angles_smoothed,
            values_smoothed,
            color_line,
            alpha=0.05)

    # ------------
    # Write outs
    # ------------
    font_additional_info = plotting_styles.font_info(size=4)

    for cnt, entry in enumerate(list_diff_max_h):
        plt.text(
            0.15,
            0 + cnt/50,
            entry,
            fontsize=2,
            fontdict=font_additional_info,
            transform=plt.gcf().transFigure)

    # ------------
    # Legend
    # ------------
    plt.legend(
        ncol=1,
        bbox_to_anchor=(0.3, -0.05),
        prop={'size': 4},
        frameon=False)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()

    plt.close()



"""Reading raw data

This file holds all functions necessary to read in information
and data to run the energy demand model.
"""
import os
import csv
import math
import logging
from collections import defaultdict
import fiona
import pandas as pd
from shapely.geometry import shape, mapping
import numpy as np
from ruamel.yaml import YAML

from energy_demand.technologies import tech_related
from energy_demand.profiles import load_profile
from energy_demand.basic import lookup_tables

def read_yaml(file_path):
    """Parse yaml config file into plain data (lists, dicts and simple values)

    Parameters
    ----------
    file_path : str
        The path of the configuration file to parse
    """
    with open(file_path, 'r') as file_handle:
        return YAML(typ='unsafe').load(file_handle)

class TechnologyData(object):
    """Class to store technology related data

    Arguments
    ---------
    fueltype : str
        Fueltype of technology
    eff_by : str, default=1
        Efficiency of technology in base year
    eff_ey : str, default=1
        Efficiency of technology in future year
    year_eff_ey : int
        Future year when eff_ey is fully realised
    eff_achieved : float
        Factor of how much of the efficienc future
        efficiency is achieved
    diff_method : float
        Differentiation method
    market_entry : int,default=2015
        Year when technology comes on the market
    tech_type : list
        Technology type
    tech_max_share : float
        Maximum theoretical fraction of how much
        this indivdual technology can contribute
        to total energy service of its enduse
    fueltypes : crit or bool,default=None
        Fueltype or criteria
    """
    def __init__(
            self,
            name=None,
            fueltype=None,
            eff_by=None,
            eff_ey=None,
            year_eff_ey=None,
            eff_achieved=None,
            diff_method=None,
            market_entry=2015,
            tech_type=None,
            tech_max_share=None,
            description=None
        ):
        self.name = name
        self.fueltype_str = fueltype
        self.fueltype_int = tech_related.get_fueltype_int(fueltype)
        self.eff_by = eff_by
        self.eff_ey = eff_ey
        self.year_eff_ey = year_eff_ey
        self.eff_achieved = eff_achieved
        self.diff_method = diff_method
        self.market_entry = market_entry
        self.tech_type = tech_type
        self.tech_max_share = tech_max_share
        self.description = description

    def set_tech_attr(self, attribute_to_set, value_to_set):
        """Set a technology attribute

        Arguments
        ----------
        attribute_to_set : str
            Attribue to set
        value_to_set : any
            Value to set
        """
        setattr(self, attribute_to_set, value_to_set)

class CapacitySwitch(object):
    """Capacity switch class for storing
    switches

    Arguments
    ---------
    enduse : str
        Enduse of affected switch
    technology_install : str
        Installed technology
    switch_yr : int
        Year until capacity installation is fully realised
    installed_capacity : float
        Installed capacity in GWh
    """
    def __init__(
            self,
            enduse,
            technology_install,
            switch_yr,
            installed_capacity,
            sector=None
        ):
        """Constructor
        """
        self.enduse = enduse
        self.technology_install = technology_install
        self.switch_yr = switch_yr
        self.installed_capacity = installed_capacity
        if not sector:
            self.sector = None
        elif isinstance(sector, str):
            self.sector = sector
        elif math.isnan(sector):
            self.sector = None
        else:
            self.sector = sector

    def update(self, name, value):
        """Update  switch

        Arguments
        ---------
        name : str
            name of attribute
        value : any
            Type of value
        """
        setattr(self, name, value)

class FuelSwitch(object):
    """Fuel switch class for storing
    switches

    Arguments
    ---------
    enduse : str
        Enduse of affected switch
    fueltype_replace : str
        Fueltype which is beeing switched from
    technology_install : str
        Installed technology
    switch_yr : int
        Year until switch is fully realised
    fuel_share_switched_ey : float
        Switched fuel share
    """
    def __init__(
            self,
            enduse=None,
            fueltype_replace=None,
            technology_install=None,
            switch_yr=None,
            fuel_share_switched_ey=None,
            sector=None
        ):
        """Constructor
        """
        self.enduse = enduse
        self.fueltype_replace = fueltype_replace
        self.technology_install = technology_install
        self.switch_yr = switch_yr
        self.fuel_share_switched_ey = fuel_share_switched_ey
        if not sector:
            self.sector = None
        elif isinstance(sector, str):
            self.sector = sector
        elif math.isnan(sector):
            self.sector = None
        else:
            self.sector = sector

    def update(self, name, value):
        """Update  switch

        Arguments
        ---------
        name : str
            name of attribute
        value : any
            Type of value
        """
        setattr(self, name, value)

class ServiceSwitch(object):
    """Service switch class for storing
    switches

    Arguments
    ---------
    enduse : str
        Enduse of affected switch
    sector : str
        Sector
    technology_install : str
        Installed technology
    service_share_ey : float
        Service share of installed technology in future year
    switch_yr : int
        Year until switch is fully realised
    """
    def __init__(
            self,
            enduse=None,
            sector=None,
            technology_install=None,
            service_share_ey=None,
            switch_yr=None
        ):
        """Constructor
        """
        self.enduse = enduse
        self.technology_install = technology_install
        self.service_share_ey = service_share_ey
        self.switch_yr = switch_yr
        if not sector:
            self.sector = None
        elif isinstance(sector, str):
            self.sector = sector
        elif math.isnan(sector):
            self.sector = None
        else:
            self.sector = sector

    def update(self, name, value):
        """Update service switch

        Arguments
        ---------
        name : str
            name of attribute
        value : any
            Type of value
        """
        setattr(self, name, value)

def read_in_results(
        path_result,
        seasons,
        model_yeardays_daytype
    ):
    """Read and post calculate results from txt files
    and store into container

    Arguments
    ---------
    path_result : str
        Paths
    seasons : dict
        seasons
    model_yeardays_daytype : dict
        Daytype of modelled yeardays
    """
    logging.info("... Reading in results")

    lookups = lookup_tables.basic_lookups()

    results_container = {}

    # -----------------
    # Read in demands
    # -----------------
    try:
        results_container['results_enduse_every_year'] = read_enduse_specific_results(
            path_result)
    except:
        pass
    try:
        print("path_result " + str(path_result))
        results_container['ed_fueltype_regs_yh'] = read_results_yh(
            path_result, 'ed_fueltype_regs_yh')
    except:
        pass
    # Read in residential demands
    try:
        results_container['residential_results'] = read_results_yh(
            path_result, 'residential_results')
    except:
        pass

    # Calculate total demand per fueltype for every hour
    try:
        tot_fueltype_yh = {}
        for year in results_container['ed_fueltype_regs_yh']:
            nr_of_fueltypes = results_container['ed_fueltype_regs_yh'][year].shape[0]
            tot_fueltype_yh[year] = np.zeros((nr_of_fueltypes, 8760))

        for year, ed_regs_yh in results_container['ed_fueltype_regs_yh'].items():
            fuel_yh = np.sum(ed_regs_yh, axis=1) #Sum across all regions
            tot_fueltype_yh[year] += fuel_yh

        results_container['tot_fueltype_yh'] = tot_fueltype_yh
    except:
        pass

    # -----------------
    # Peak calculations
    # -----------------
    try:
        results_container['ed_peak_h'] = {}
        results_container['ed_peak_regs_h'] = {}

        for year, ed_fueltype_reg_yh in results_container['ed_fueltype_regs_yh'].items():
            results_container['ed_peak_h'][year] = {}
            results_container['ed_peak_regs_h'][year] = {}

            for fueltype_int, ed_reg_yh in enumerate(ed_fueltype_reg_yh):
                fueltype_str = tech_related.get_fueltype_str(lookups['fueltypes'], fueltype_int)

                # Calculate peak per fueltype for all regions (ed_reg_yh= np.array(fueltype, reg, yh))
                all_regs_yh = np.sum(ed_reg_yh, axis=0)    # sum regs
                peak_h = np.max(all_regs_yh)               # select max of 8760 h
                results_container['ed_peak_h'][year][fueltype_str] = peak_h
                results_container['ed_peak_regs_h'][year][fueltype_str] = np.max(ed_reg_yh, axis=1)

        # -------------
        # Load factors
        # -------------
        results_container['reg_load_factor_y'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_y"))
        results_container['reg_load_factor_yd'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_yd"))

        # -------------
        # Post-calculations
        # -------------
        # Calculate average per season and fueltype for every fueltype
        results_container['av_season_daytype_cy'], results_container['season_daytype_cy'] = calc_av_per_season_fueltype(
            results_container['ed_fueltype_regs_yh'],
            seasons,
            model_yeardays_daytype)

        '''results_container['load_factor_seasons'] = {}
        results_container['load_factor_seasons']['winter'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_winter"))
        results_container['load_factor_seasons']['spring'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_spring"))
        results_container['load_factor_seasons']['summer'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_summer"))
        results_container['load_factor_seasons']['autumn'] = read_lf_y(
            os.path.join(path_result, "result_reg_load_factor_autumn"))'''
    except:
        pass
    logging.info("... Reading in results finished")
    return results_container

def calc_av_per_season_fueltype(results_every_year, seasons, model_yeardays_daytype):
    """Calculate average demand per season and fueltype for every fueltype

    Arguments
    ---------
    results_every_year : dict
        Results for every year
    seasons : dict
        Seasons
    model_yeardays_daytype : list
        Daytype of modelled days

    Returns
    -------
    av_season_daytype_cy :
        Average demand per season and daytype
    season_daytype_cy :
        Demand per season and daytpe
    """
    av_season_daytype_cy = defaultdict(dict)
    season_daytype_cy = defaultdict(dict)

    for year, fueltypes_data in results_every_year.items():

        for fueltype, reg_fuels in enumerate(fueltypes_data):

            # Summarise across regions
            tot_all_reg_fueltype = np.sum(reg_fuels, axis=0)

            tot_all_reg_fueltype_reshape = tot_all_reg_fueltype.reshape((365, 24))

            calc_av, calc_lp = load_profile.calc_av_lp(
                tot_all_reg_fueltype_reshape,
                seasons,
                model_yeardays_daytype)

            av_season_daytype_cy[year][fueltype] = calc_av
            season_daytype_cy[year][fueltype] = calc_lp

    return dict(av_season_daytype_cy), dict(season_daytype_cy)

def read_results_yh(path_to_folder, name_of_folder):
    """Read results

    Arguments
    ---------
    fueltypes_nr : int
        Number of fueltypes
    reg_nrs : int
        Number of regions
    path_to_folder : str
        Path to folder

    Returns
    -------
    results = dict
        Results
    """
    results = {}

    path_to_folder = os.path.join(path_to_folder, name_of_folder)

    all_txt_files_in_folder = os.listdir(path_to_folder)

    for file_path in all_txt_files_in_folder:
        try:
            path_file_to_read = os.path.join(path_to_folder, file_path)
            file_path_split = file_path.split("__")
            year = int(file_path_split[1][:-4])

            results[year] = np.load(path_file_to_read)

        except IndexError:
            pass #path is a folder and not a file

    return results

def read_max_results(path):
    """Read max results

    Arguments
    ---------
    path : str
        Path to folder
    """
    results = {}
    all_txt_files_in_folder = os.listdir(path)

    # Iterate files
    for file_path in all_txt_files_in_folder:
        path_file_to_read = os.path.join(path, file_path)
        file_path_split = file_path.split("__")
        year = int(file_path_split[1])

        # Add year if not already exists
        results[year] = np.load(path_file_to_read)

    return results

def read_enduse_specific_results(path_to_folder):
    """Read enduse specific results

    Arguments
    ---------
    path_to_folder : str
        Folder path
    """
    results = defaultdict(dict)

    path_results = os.path.join(
        path_to_folder,
        "enduse_specific_results")

    all_txt_files_in_folder = os.listdir(path_results)

    for file_path in all_txt_files_in_folder:
        path_file_to_read = os.path.join(path_results, file_path)
        file_path_split = file_path.split("__")

        if file_path_split[-1] == '.txt':
            pass
        else:
            enduse = file_path_split[1]
            year = int(file_path_split[2])

            results[year][enduse] = np.load(path_file_to_read)

    return dict(results)

def read_fuel_ss(path_to_csv, fueltypes_nr):
    """This function reads in base_data_CSV all fuel types

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    fueltypes_nr : str
        Nr of fueltypes

    Returns
    -------
    fuels : dict
        Fuels per enduse
    sectors : list
        Service sectors
    enduses : list
        Service enduses
    
    Info of categories
    ------------------
    https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/565748/BEES_overarching_report_FINAL.pdf
    """
    lookups = lookup_tables.basic_lookups()
    fueltypes_lu = lookups['fueltypes']

    rows_list = []
    fuels = {}
    try:
        with open(path_to_csv, 'r') as csvfile:
            rows = csv.reader(csvfile, delimiter=',')
            headings = next(rows) # Skip row
            _secondline = next(rows) # Skip row

            # All sectors
            sectors = set([])
            for sector in _secondline[1:]: #skip fuel ID:
                sectors.add(sector)

            # All enduses
            enduses = set([])
            for enduse in headings[1:]: #skip fuel ID:
                enduses.add(enduse)

            # Initialise dict
            for enduse in enduses:
                fuels[enduse] = {}
                for sector in sectors:
                    fuels[enduse][sector] = np.zeros((fueltypes_nr), dtype="float")

            for row in rows:
                rows_list.append(row)

            for row in rows_list:
                fueltype_str = row[0]
                fueltype_int = fueltypes_lu[fueltype_str]
                for cnt, entry in enumerate(row[1:], 1):
                    enduse = headings[cnt]
                    sector = _secondline[cnt]
                    fuels[enduse][sector][fueltype_int] += float(entry)
    except ValueError:
        raise Exception(
            "The service sector fuel could not be loaded. Check if empty cells.")

    return fuels, list(sectors), list(enduses)

def read_load_shapes_tech(path_to_csv):
    """This function reads in csv technology shapes

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    """
    load_shapes_dh = {}

    with open(path_to_csv, 'r') as csvfile:
        rows = csv.reader(csvfile, delimiter=',')
        headings = next(rows) # Skip first row

        for row in rows:
            dh_shape = np.zeros((24), dtype="float")
            for cnt, row_entry in enumerate(row[1:], 1):
                dh_shape[int(headings[cnt])] = float(row_entry)

            load_shapes_dh[str(row[0])] = dh_shape

    return load_shapes_dh

def service_switch(df_service_switches):
    """This function reads in service assumptions from csv file,
    tests whether the maximum defined switch is larger than
    possible for a technology.

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    technologies : list
        All technologies

    Returns
    -------
    enduse_tech_ey_p : dict
        Technologies per enduse for endyear in p
    service_switches : dict
        Service switches

    Notes
    -----
    The base year service shares are generated from technology stock definition

    Info
    -----
    The following attributes need to be defined for a service switch.

        Attribute                   Description
        ==========                  =========================
        enduse                      [str]   Enduse affected by switch
        tech                        [str]   Technology
        switch_yr                   [int]   Year until switch is fully realised
        service_share_ey            [str]   Service share of 'tech' in 'switch_yr'
        sector                      [str]   Optional sector specific info where switch applies
    """
    service_switches = []
    default_parameter = 999.0   #default parameter

    for i in df_service_switches.index:
        enduse = df_service_switches.at[i, 'enduses_service_switch']
        tech = df_service_switches.at[i, 'tech']
        service_share_ey = df_service_switches.at[i, 'switches_service']
        switch_yr = df_service_switches.at[i, 'end_yr']
        sector = df_service_switches.at[i, 'sector']

        if sector == 'None':
            sector = None

        if float(service_share_ey) == default_parameter:
            pass
        else:
            service_switches.append(
                ServiceSwitch(
                    enduse=str(enduse),
                    technology_install=str(tech),
                    service_share_ey=float(service_share_ey),
                    switch_yr=float(switch_yr),
                    sector=sector))

    return service_switches

def read_fuel_switches(
        path_to_csv,
        enduses,
        fueltypes,
        technologies,
        base_yr=2015
    ):
    """This function reads in from CSV file defined fuel
    switch assumptions

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    enduses : dict
        Endues per submodel
    fueltypes : dict
        Look-ups
    technologies : dict
        Technologies

    Returns
    -------
    dict_with_switches : dict
        All assumptions about fuel switches provided as input


    Info
    -----
    The following attributes need to be defined for a fuel switch.

        Attribute                   Description
        ==========                  =========================
        enduse                      [str]   Enduse affected by switch
        fueltype_replace            [str]   Fueltype to be switched from
        technology_install          [str]   Technology which is installed
        switch_yr                   [int]   Year until switch is fully realised
        fuel_share_switched_ey      [float] Share of fuel which is switched until switch_yr
        sector                      [str]   Optional sector specific info where switch applies
                                            If field is empty the switch is across all sectors
    """
    fuel_switches = []

    if os.path.isfile(path_to_csv):

        raw_csv_file = pd.read_csv(path_to_csv)

        for index, row in raw_csv_file.iterrows():
            fuel_switches.append(
                FuelSwitch(
                    enduse=str(row['enduse']),
                    fueltype_replace=fueltypes[str(row['fueltype_replace'])],
                    technology_install=str(row['technology_install']),
                    switch_yr=float(row['switch_yr']),
                    fuel_share_switched_ey=float(row['fuel_share_switched_ey']),
                    sector=row['sector']))

        # -------
        # Testing
        #
        # Test if more than 100% per fueltype is switched or more than
        # than theoretically possible per technology
        # --------
        # Testing wheter the provided inputs make sense
        for obj in fuel_switches:
            if obj.fuel_share_switched_ey == 0:
                raise Exception(
                    "Input error: The share of switched fuel must be > 0. Delete {} from input".format(
                        obj.technology_install))

            for obj_iter in fuel_switches:

                # Test if lager than maximum defined technology diffusion (L)
                if obj_iter.fuel_share_switched_ey > technologies[obj_iter.technology_install].tech_max_share:
                    raise Exception(
                        "Configuration Error: More service provided for tech '{}' in enduse '{}' than max possible".format(
                            obj_iter.enduse, obj_iter.technology_install))

                if obj_iter.fuel_share_switched_ey > 1.0:
                    raise Exception(
                        "Configuration Error: The fuel switches are > 1.0 for enduse {} and fueltype {}".format(
                            obj.enduse, obj.fueltype_replace))

            if obj.switch_yr <= base_yr:
                raise Exception("Configuration Error of fuel switch: switch_yr must be in the future")

        # Test whether defined enduse exist
        for obj in fuel_switches:
            if obj.enduse in enduses['service'] or obj.enduse in enduses['residential'] or obj.enduse in enduses['industry']:
                pass
            else:
                raise Exception(
                    "Input Error: The defined enduse '{}' to switch fuel from is not defined...".format(
                        obj.enduse))
    else:
        pass

    return fuel_switches

def read_technologies(path_to_csv):
    """Read in technology definition csv file. Append
    for every technology type a 'placeholder_tech'.

    Arguments
    ----------
    path_to_csv : str
        Path to csv file

    Returns
    -------
    dict_technologies : dict
        All technologies and their assumptions provided as input
    dict_tech_lists : dict
        List with technologies. The technology type
        is defined in the technology input file. A placeholder technology
        is added for every list in order to allow that a generic
        technology type can be added for every enduse

    Info
    -----
    The following attributes need to be defined for implementing
    a technology.

        Attribute                   Description
        ==========                  =========================
        technology                  [str]   Name of technology
        fueltype                    [str]   Fueltype of technology
        eff_by                      [float] Efficiency in base year
        eff_ey                      [float] Efficiency in future end year
        year_eff_ey	                [int]   Future year where efficiency is fully reached
        eff_achieved                [float] Factor of how much of the efficiency
                                            is achieved (overwritten by scenario input)
                                            This is set to 1.0 as default for initial
                                            technology class generation
        diff_method	market_entry    [int]   Year of market entry of technology
        tech_list                   [str]   Definition of to which group
                                            of technologies a technology belongs
        tech_max_share              [float] Maximum share of technology related
                                            energy service which can be reached in theory
        description                 [str]   Optional technology description
    """
    dict_technologies = {}
    dict_tech_lists = {}

    raw_csv_file = pd.read_csv(path_to_csv)

    for index, row in raw_csv_file.iterrows():
        dict_technologies[str(row['technology'])] = TechnologyData(
            name=str(row['technology']),
            fueltype=str(row['fueltype']),
            eff_by=float(row['efficiency in base year']),
            eff_ey=float(row['efficiency in future year']),
            year_eff_ey=float(row['year when efficiency is fully realised']),
            eff_achieved=1.0, # Set to one as default
            diff_method=str(row['diffusion method (sigmoid or linear)']),
            market_entry=float(row['market_entry']),
            tech_type=str(row['technology type']),
            tech_max_share=float(row['maximum theoretical service share of technology']),
            description=str(row['description']))
        try:
            dict_tech_lists[row['technology type']].append(row['technology'])
        except KeyError:
            dict_tech_lists[row['technology type']] = [row['technology']]

    # Add placeholder technology to all tech_lists
    for tech_list in dict_tech_lists.values():
        tech_list.append('placeholder_tech')

    return dict_technologies, dict_tech_lists

def read_fuel_rs(path_to_csv):
    """This function reads in base_data_CSV all fuel types

    (first row is fueltype, subkey), header is appliances

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    _dt : str
        Defines dtype of array to be read in (takes float)

    Returns
    -------
    fuels : dict
        Residential fuels
    enduses : list
        Residential end uses

    Notes
    -----
    the first row is the fuel_ID
    The header is the sub_key
    """
    dummy_sector = None
    sectors = [dummy_sector]

    fuels = {}

    # Read csv
    raw_csv_file = pd.read_csv(path_to_csv)

    # Replace NaN with " " values
    raw_csv_file = raw_csv_file.fillna(0)

    # Enduses
    enduses = list(raw_csv_file.columns[1:].values) #skip fuel_id

    # Replace str fueltypes with int fueltypes
    raw_csv_file['fuel_id'] = raw_csv_file['fuel_id'].apply(tech_related.get_fueltype_int)

    # Iterate columns and convert to array
    for enduse in raw_csv_file.columns[1:]: # skip for column
        fuels[enduse] = {}
        fuels[enduse][dummy_sector] = raw_csv_file[enduse].values

    return fuels, sectors, list(enduses)

def read_fuel_is(path_to_csv, fueltypes_nr):
    """This function reads in base_data_CSV all fuel types

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    fueltypes_nr : int
        Number of fueltypes

    Returns
    -------
    fuels : dict
        Industry fuels
    sectors : list
        Industral sectors
    enduses : list
        Industrial enduses

    Info
    ----
    Source: User Guide Energy Consumption in the UK
            https://www.gov.uk/government/uploads/system/uploads/attach
            ment_data/file/573271/ECUK_user_guide_November_2016_final.pdf

            https://unstats.un.org/unsd/cr/registry/regcst.asp?Cl=27

            http://ec.europa.eu/eurostat/ramon/nomenclatures/
            index.cfm?TargetUrl=LST_NOM_DTL&StrNom=NACE_REV2&StrLanguageCode=EN&IntPcKey=&StrLayoutCode=

    High temperature processes
    =============================
    High temperature processing dominates energy consumption in the iron and steel,
    non-ferrous metal, bricks, cement, glass and potteries industries. This includes
        - coke ovens
        - blast furnaces and other furnaces
        - kilns and
        - glass tanks.

    Low temperature processes
    =============================
    Low temperature processes are the largest end use of energy for the food, drink
    and tobacco industry. This includes:
        - process heating and distillation in the chemicals sector;
        - baking and separation processes in food and drink;
        - pressing and drying processes, in paper manufacture;
        - and washing, scouring, dyeing and drying in the textiles industry.

    Drying/separation
    =============================
    Drying and separation is important in paper-making while motor processes are used
    more in the manufacture of chemicals and chemical products than in any other
    individual industry.

    Motors
    =============================
    This includes pumping, fans and machinery drives.

    Compressed air
    =============================
    Compressed air processes are mainly used in the publishing, printing and
    reproduction of recorded media sub-sector.

    Lighting
    =============================
    Lighting (along with space heating) is one of the main end uses in engineering
    (mechanical and electrical engineering and vehicles industries).

    Refrigeration
    =============================
    Refrigeration processes are mainly used in the chemicals and food and drink
    industries.

    Space heating
    =============================
    Space heating (along with lighting) is one of the main end uses in engineering
    (mechanical and electrical engineering and vehicles industries).

    Other
    =============================

    -----------------------
    Industry classes from BEIS
    -----------------------

    SIC 2007    Name
    --------    ------
    08	        Other mining and quarrying
    10	        Manufacture of food products
    11	        Manufacture of beverages
    12	        Manufacture of tobacco products
    13	        Manufacture of textiles
    14	        Manufacture of wearing apparel
    15	        Manufacture of leather and related products
    16	        Manufacture of wood and of products of wood and cork, except furniture; manufacture of articles of straw and plaiting materials
    17	        Manufacture of paper and paper products
    18	        Printing and publishing of recorded media and other publishing activities
    20	        Manufacture of chemicals and chemical products
    21	        Manufacture of basic pharmaceutical products and pharmaceutical preparations
    22	        Manufacture of rubber and plastic products
    23	        Manufacture of other non-metallic mineral products
    24	        Manufacture of basic metals
    25	        Manufacture of fabricated metal products, except machinery and equipment
    26	        Manufacture of computer, electronic and optical products
    27	        Manufacture of electrical equipment
    28	        Manufacture of machinery and equipment n.e.c.
    29	        Manufacture of motor vehicles, trailers and semi-trailers
    30	        Manufacture of other transport equipment
    31	        Manufacture of furniture
    32	        Other manufacturing
    36	        Water collection, treatment and supply
    38	        Waste collection, treatment and disposal activities; materials recovery
    """
    rows_list = []
    fuels = {}

    '''# Read csv
    raw_csv_file = pd.read_csv(path_to_csv)

    # Replace NaN with " " values
    raw_csv_file = raw_csv_file.fillna(0)

    # Enduses
    enduses = list(raw_csv_file.columns.values)'''


    with open(path_to_csv, 'r') as csvfile:
        rows = csv.reader(csvfile, delimiter=',')
        headings = next(rows)
        _secondline = next(rows)

        # All sectors
        enduses = set([])
        for enduse in headings[1:]:
            if enduse is not '':
                enduses.add(enduse)

        # All enduses
        sectors = set([])
        for row in rows:
            rows_list.append(row)
            sectors.add(row[0])

        # Initialise dict
        for enduse in enduses:
            fuels[enduse] = {}
            for sector in sectors:

                fuels[str(enduse)][str(sector)] = np.zeros(
                    (fueltypes_nr), dtype="float")

        for row in rows_list:
            sector = row[0]
            for position, entry in enumerate(row[1:], 1): # Start with position 1

                if entry != '':
                    enduse = str(headings[position])
                    fueltype = _secondline[position]
                    fueltype_int = tech_related.get_fueltype_int(fueltype)
                    fuels[enduse][sector][fueltype_int] += float(row[position])

    return fuels, list(sectors), list(enduses)

def read_lf_y(result_path):
    """Read load factors from .npy file

    Arguments
    ----------
    result_path : str
        Path

    Returns
    -------
    results : dict
        Annual results
    """
    results = {}

    all_txt_files_in_folder = os.listdir(result_path)

    for file_path in all_txt_files_in_folder:
        path_file_to_read = os.path.join(result_path, file_path)
        file_path_split = file_path.split("__")
        year = int(file_path_split[1])

        results[year] = np.load(path_file_to_read)

    return results

def read_scenaric_population_data(result_path):
    """Read population data

    Arguments
    ---------
    result_path : str
        Path

    Returns
    -------
    results : dict
        Population, {year: np.array(fueltype, regions)}

    """
    results = {}

    all_txt_files_in_folder = os.listdir(result_path)

    for file_path in all_txt_files_in_folder:
        path_file_to_read = os.path.join(result_path, file_path)
        file_path_split = file_path.split("__")
        year = int(file_path_split[1])

        # Add year if not already exists
        results[year] = np.load(path_file_to_read)

    return results

def read_capacity_switch(path_to_csv, base_yr=2015):
    """This function reads in service assumptions
    from csv file

    Arguments
    ----------
    path_to_csv : str
        Path to csv file

    Returns
    -------
    service_switches : dict
        Service switches which implement the defined capacity installation

    Info
    -----
    The following attributes need to be defined for a capacity switch.

        Attribute                   Description
        ==========                  =========================
        enduse                      [str]   Enduse affected by switch
        tech                        [str]   Technology installed
        switch_yr                   [int]   Year until switch is fully realised
        installed_capacity          [float] Installed total capacity in GWh
        sector                      [str]   Optional sector specific info where switch applies
                                            If field is empty the switch is across all sectors
    """
    service_switches = []

    if os.path.isfile(path_to_csv):

        # Read switches
        raw_csv_file = pd.read_csv(path_to_csv)

        # Iterate rows
        for _, row in raw_csv_file.iterrows():

            service_switches.append(
                CapacitySwitch(
                    enduse=str(row['enduse']),
                    technology_install=str(row['technology_install']),
                    switch_yr=float(row['swich_yr']),
                    installed_capacity=float(row['installed_capacity']),
                    sector=row['sector']))

        # Testing
        for obj in service_switches:
            if obj.switch_yr <= base_yr:
                raise Exception("Input Error capacity switch: switch_yr must be in the future")

    else:
        pass
    return service_switches

def read_floor_area_virtual_stock(path_to_csv, f_mixed_floorarea=0.5):
    """Read in floor area from csv file for every LAD
    to generate virtual building stock.

    This file is obainted from Newcastle

    Arguments
    ---------
    path_to_csv : str
        Path to csv file
    f_mixed_floorarea : float
        Factor to assign mixed floor area

    Returns
    -------
    res_floorarea : dict
        Residential floor area per region
    non_res_floorarea : dict
        Non residential floor area per region

    Info
    -----
    *   The mixed floor area (residential and non residential) is distributed
        according to `f_mixed_floorarea`.

    Attributes from data from Newcastle
    ===================================
    (1) Commercial_General
    (2) Primary_Industry
    (3) Public_Services
    (4) Education
    (5) Hospitality
    (6) Community_Arts_Leisure
    (7) Industrial
    (8) Healthcare
    (9) Office
    (10) Retail
    (11) Transport_and_Storage
    (12) Residential
    (13) Military
    """
    # Redistribute the mixed enduse
    p_mixed_no_resid = 1 - f_mixed_floorarea

    # Second Mail from Craig
    res_floorarea, non_res_floorarea, floorarea_mixed = {}, {}, {}

    building_count_service = {}
    for i in range(1, 15):
        building_count_service[i] = {}

    with open(path_to_csv, 'r') as csvfile:
        rows = csv.reader(csvfile, delimiter=',')
        headings = next(rows)

        for row in rows:
            geo_name = str.strip(row[get_position(headings, 'lad')])

            if row[get_position(headings, 'res_bld_floor_area')] == 'null':
                # Not data or faulty data
                pass
            else:
                res_floorarea[geo_name] = float(row[get_position(headings, 'res_bld_floor_area')])

            if row[get_position(headings, 'nonres_bld_floor_area')] == 'null':
                # Not data or faulty data
                pass
            else:
                non_res_floorarea[geo_name] = float(row[get_position(headings, 'nonres_bld_floor_area')])

            if row[get_position(headings, 'mixeduse_bld_floor_area')] == 'null':
                # Not data or faulty data
                pass
            else:
                floorarea_mixed[geo_name] = float(row[get_position(headings, 'mixeduse_bld_floor_area')])

                # Distribute mixed floor area
                non_res_from_mixed = floorarea_mixed[geo_name] * p_mixed_no_resid
                res_from_mixed = floorarea_mixed[geo_name] * f_mixed_floorarea

                # Add
                res_floorarea[geo_name] += res_from_mixed
                non_res_floorarea[geo_name] += non_res_from_mixed

            # ---------------------------------------
            # Read building count for service sector
            # ---------------------------------------
            building_1 = float(row[get_position(headings, 'building_type_count_1')])
            building_2 = float(row[get_position(headings, 'building_type_count_2')])
            building_3 = float(row[get_position(headings, 'building_type_count_3')])
            building_4 = float(row[get_position(headings, 'building_type_count_4')])
            building_5 = float(row[get_position(headings, 'building_type_count_5')])
            building_6 = float(row[get_position(headings, 'building_type_count_6')])
            building_7 = float(row[get_position(headings, 'building_type_count_7')])
            building_8 = float(row[get_position(headings, 'building_type_count_8')])
            building_9 = float(row[get_position(headings, 'building_type_count_9')])
            building_10 = float(row[get_position(headings, 'building_type_count_10')])
            building_11 = float(row[get_position(headings, 'building_type_count_11')])
            building_12 = float(row[get_position(headings, 'building_type_count_12')])
            building_13 = float(row[get_position(headings, 'building_type_count_13')])

            building_count_service[1][geo_name] = building_1
            building_count_service[2][geo_name] = building_2
            building_count_service[3][geo_name] = building_3
            building_count_service[4][geo_name] = building_4
            building_count_service[5][geo_name] = building_5
            building_count_service[6][geo_name] = building_6
            building_count_service[7][geo_name] = building_7
            building_count_service[8][geo_name] = building_8
            building_count_service[9][geo_name] = building_9
            building_count_service[10][geo_name] = building_10
            building_count_service[11][geo_name] = building_11
            building_count_service[12][geo_name] = building_12
            building_count_service[13][geo_name] = building_13

            # Create Other category and buildings
            building_count_service[14][geo_name] = int(
                building_1 + building_2 + building_3 +
                building_4 + building_5 + building_6 +
                building_7 + building_8 + building_9 +
                building_10 + building_11 + building_12 +
                building_13)

    return res_floorarea, non_res_floorarea, building_count_service

def get_position(headings, name):
    """Get position of an entry in a list

    Arguments
    ---------
    headings : list
        List with names
    name : str
        Name of entry to find

    Returns
    -------
    position : int
        Position in list
    """
    return headings.index(name)

def read_np_array_from_txt(path_file_to_read):
    """Read np array from textfile

    Arguments
    ---------
    path_file_to_read : str
        File to path with stored array

    Return
    ------
    txt_array : array
        Array containing read text
    """
    txt_array = np.loadtxt(path_file_to_read, delimiter=',')

    return txt_array

def get_region_selection(path_to_csv):
    """Read region names in a csv

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    """
    regions = []

    with open(path_to_csv, 'r') as csvfile:
        rows = csv.reader(csvfile, delimiter=',')
        _headings = next(rows)

        for row in rows:
            regions.append(row[0])

    return regions

def get_region_names(path):
    '''Returns names of shapes within a shapefile
    '''
    with fiona.open(path, 'r') as source:
        return [elem['properties']['name'] for elem in source]

def get_region_centroids(path):
    '''Returns centroids of shapes within a shapefile
    '''
    with fiona.open(path, 'r') as source:
        geoms = [elem for elem in source]

    for geom in geoms:
        my_shape = shape(geom['geometry'])
        geom['geometry'] = mapping(my_shape.centroid)

    return geoms

def get_region_objects(path):
    '''Returns shape objects within a shapefile
    '''
    with fiona.open(path, 'r') as source:
        return [elem for elem in source]

def load_full_paramter_values(file_path):
    """
    """
    # READ csv file
    # "region", "year", "value", "interval"
    gp_file = pd.read_csv(file_path)

    return gp_file



"""Script to convert fuel to energy service
"""
import numpy as np
from collections import defaultdict

from energy_demand.basic import lookup_tables
from energy_demand.technologies import tech_related
from energy_demand.initalisations import helpers

def init_nested_dict_brackets(first_level_keys, second_level_keys):
    """Initialise a nested dictionary with two levels

    Arguments
    ----------
    first_level_keys : list
        First level data
    second_level_keys : list
        Data to add in nested dict

    Returns
    -------
    nested_dict : dict
        Nested 2 level dictionary
    """
    nested_dict = {}

    for first_level_key in first_level_keys:
        nested_dict[first_level_key] = {}
        for second_level_key in second_level_keys:
            nested_dict[first_level_key][second_level_key] = {}

    return nested_dict

def init_nested_dict_zero(sector, first_level_keys, second_level_keys):
    """Initialise a nested dictionary with two levels

    Arguments
    ----------
    first_level_keys : list
        First level data
    second_level_keys : list
        Data to add in nested dict

    Returns
    -------
    nested_dict : dict
        Nested 2 level dictionary
    """
    nested_dict = {}

    for first_level_key in first_level_keys:
        nested_dict[first_level_key] = {}
        nested_dict[first_level_key][sector] = {}
        for second_level_key in second_level_keys:
            nested_dict[first_level_key][sector][second_level_key] = 0

    return nested_dict

def sum_2_level_dict(two_level_dict):
    """Sum all entries in a two level dict

    Arguments
    ----------
    two_level_dict : dict
        Nested dict

    Returns
    -------
    tot_sum : float
        Number of all entries in nested dict
    """
    tot_sum = 0
    for j in two_level_dict.values():
        tot_sum += sum(j.values())

    return tot_sum

def sum_fuel_enduse_sectors(data_enduses, enduses):
    """Aggregate fuel for all sectors according to enduse

    Arguments
    --------
    data_enduses : dict
        Fuel per enduse
    enduses : list
        Enduses
    nr_fueltypes : int
        Number of fuetlypes

    Returns
    -------
    aggregated_fuel_enduse : dict
        Arregated fuel per enduse ({enduse: np.array(fueltype)})
    """
    aggregated_fuel_enduse = {}

    for enduse in enduses:
        aggregated_fuel_enduse[enduse] = sum(data_enduses[enduse].values())

    return aggregated_fuel_enduse

def get_s_fueltype_tech(
        enduses,
        fuel_p_tech_by,
        fuels,
        technologies,
        sector=False
    ):
    """Calculate total energy service fractions per technology.
    This calculation converts fuels into energy services (e.g. heating
    for fuel into heat demand) and then calculates how much an invidual
    technology contributes as a fraction of total energy service demand.

    This is calculated to determine how much the technology
    has already diffused up to the base year to define the
    first point on the sigmoid technology diffusion curve.

    Arguments
    ----------
    enduses : dict
        Enduses
    fuel_p_tech_by : dict
        Assumed fraction of fuel for each technology within a fueltype
    fuels : array
        Base year fuel demand
    technologies : object
        Technology of base year (region dependent)

    Return
    ------
    s_tech_by_p : dict
        Percentage of total energy service per technology for base year
    s_fueltype_by_p : dict
        Percentage of energy service per fueltype
    """
    fueltypes =  lookup_tables.basic_lookups()['fueltypes']

    service = init_nested_dict_brackets(fuels, fueltypes.values()) # Energy service per technology for base year
    s_tech_by_p = helpers.init_dict_brackets(fuels) # Percentage of total energy service per technology for base year
    s_fueltype_by_p = init_nested_dict_zero(sector, s_tech_by_p.keys(), range(len(fueltypes))) # Percentage of service per fueltype

    for enduse in enduses:

        # Depending if sector or not sector specific
        fuel = fuels[enduse][sector]

        for fueltype, fuel_fueltype in enumerate(fuel):
            tot_s_fueltype = 0

            # Initialise
            for tech in fuel_p_tech_by[enduse][sector][fueltype]:
                service[enduse][fueltype][tech] = 0

            # Iterate technologies to calculate share of energy service depending on fuel and efficiencies
            for tech, fuel_alltech_by in fuel_p_tech_by[enduse][sector][fueltype].items():

                # Fuel share based on defined shares within fueltype (share of fuel * total fuel)
                fuel_tech = fuel_alltech_by * fuel_fueltype

                # Get efficiency for base year
                if technologies[tech].tech_type == 'heat_pump':
                    eff_tech = tech_related.eff_heat_pump(
                        temp_diff=10,
                        efficiency_intersect=technologies[tech].eff_by)
                else:
                    eff_tech = technologies[tech].eff_by

                # Energy service: Service == Fuel of technoloy * efficiency
                s_fueltype_tech = fuel_tech * eff_tech

                # Add energy service demand
                service[enduse][fueltype][tech] += s_fueltype_tech

                # Total energy service demand within a fueltype
                tot_s_fueltype += s_fueltype_tech

            # Calculate percentage of service enduse within fueltype
            for tech in fuel_p_tech_by[enduse][sector][fueltype]:
                if tot_s_fueltype == 0: # No fuel in this fueltype
                    s_fueltype_by_p[enduse][sector][fueltype] += 0
                else:
                    s_fueltype_by_p[enduse][sector][fueltype] += service[enduse][fueltype][tech]

        # Calculate percentage of service of all technologies
        total_s = sum_2_level_dict(service[enduse])

        # Percentage of energy service per technology
        for fueltype, technology_s_enduse in service[enduse].items():
            for technology, s_tech in technology_s_enduse.items():

                with np.errstate(divide='ignore'):

                    if total_s == 0:
                        s_p_tech = 0
                    else:
                        # Calculate service
                        s_p_tech = s_tech / total_s

                     # Do not add dummy technology with zero service
                    if technology == 'placeholder_tech' and s_p_tech == 0:
                        pass
                    else:
                        try:
                            s_tech_by_p[enduse][technology] += s_p_tech
                        except:
                            s_tech_by_p[enduse][technology] = s_p_tech

        # Convert service per enduse
        for fueltype in s_fueltype_by_p[enduse][sector]:
            with np.errstate(divide='ignore'):
                if total_s == 0:
                    s_fueltype_by_p[enduse][sector][fueltype] = 0
                else:
                    s_fueltype_by_p[enduse][sector][fueltype] = s_fueltype_by_p[enduse][sector][fueltype] / total_s

    # Test if the energy service for all technologies is 100%
    # Test if within fueltype always 100 energy service
    for enduse, s_p_techs in s_tech_by_p.items():
        sum_enduse = sum(s_p_techs.values())
        if round(sum_enduse, 2) != 0: # if total is zero, no demand provided
            assert round(sum_enduse, 2) == 1.0

    return s_tech_by_p, s_fueltype_by_p



"""Funtions related to the diffusion of technologies
"""
import math
import numpy as np

def linear_diff(base_yr, curr_yr, value_start, value_end, yr_until_changed):
    """Calculate a linear diffusion for a current year. If
    the current year is identical to the base year, the
    start value is returned

    Arguments
    ----------
    base_yr : int
        The year of the current simulation
    curr_yr : int
        The year of the current simulation
    value_start : float
        Fraction of population served with fuel_enduse_switch in base year
    value_end : float
        Fraction of population served with fuel_enduse_switch in end year
    yr_until_changed : str
        Year until changed is fully implemented

    Returns
    -------
    fract_cy : float
        The fraction in the simulation year
    """
    # Total number of simulated years
    sim_years = yr_until_changed - base_yr  + 1

    if curr_yr == base_yr or sim_years == 0 or value_end == value_start:
        fract_cy = value_start
    else:
        #-1 because in base year no change
        fract_cy = ((value_end - value_start) / (sim_years - 1)) * (curr_yr - base_yr) + value_start

    return fract_cy

def sigmoid_function(x_value, l_value, midpoint, steepness):
    """Sigmoid function used for fitting and plotting.

    Arguments
    ---------
    x_value : float
        X-Value
    l_value : float
        The curv'es maximum value
    midpoint : float
        The midpoint x-value of the sigmoid's midpoint
    steepness : dict
        The steepness of the curve

    Return
    ------
    y-value : float
        Y-Value

    Warning
    -------
    Because 2000 is substracted, the start year can not be before 2001.
    """
    return l_value / (1 + np.exp(-steepness * ((x_value - 2000.0) - midpoint)))

def sigmoid_diffusion(
        base_yr,
        curr_yr,
        end_yr,
        sig_midpoint,
        sig_steepness
    ):
    """Calculates a sigmoid diffusion path of a lower to a higher value with
    assumed saturation at the end year

    Arguments
    ----------
    base_yr : int
        Base year of simulation period
    curr_yr : int
        The year of the current simulation
    end_yr : int
        The year a fuel_enduse_switch saturaes
    sig_midpoint : float
        Mid point of sigmoid diffusion function can be used to shift
        curve to the left or right (standard value: 0)
    sig_steepness : float
        Steepness of sigmoid diffusion function The steepness of the
        sigmoid curve (standard value: 1)

    Returns
    -------
    cy_p : float
        The fraction of the diffusion in the current year

    Note
    ----
    It is always assuemed that for the simulation year the share is
    replaced with technologies having the efficencies of the current year.
    For technologies which get replaced fast (e.g. lightbulb) this
    is corret assumption, for longer lasting technologies, this is
    more problematic (in this case, over every year would need to be iterated
    and calculate share replaced with efficiency of technology in each year).

    Always returns positive value. Needs to be considered for changes in negative
    """
    if curr_yr == base_yr:
        return float(0)
    elif curr_yr == end_yr:
        return float(1)
    else:
        # Translates simulation year on the sigmoid graph reaching from -6 to +6 (x-value)
        if end_yr == base_yr:
            y_trans = 5.0
        else:
            y_trans = -5.0 + (10.0 / (end_yr - base_yr)) * (curr_yr - base_yr)

        # Get a value between 0 and 1 (sigmoid curve ranging from 0 to 1)
        cy_p = 1.0 / (1 + math.exp(-1 * sig_steepness * (y_trans - sig_midpoint)))

        return float(cy_p)



"""testing
"""
from energy_demand.profiles import hdd_cdd
import numpy as np

def calc_weekend_corr_f():
    """Testing
    """
    wkend_factor = 0.5
    model_yeardays_daytype = ['working_day', 'holiday']
    result = hdd_cdd.calc_weekend_corr_f(
        model_yeardays_daytype=model_yeardays_daytype,
        wkend_factor=wkend_factor)

    assert result[0] == 1.0
    assert result[1] == wkend_factor

def test_effective_temps():
    """Testing
    """
    temp_yh = np.zeros((365, 24))

    for day in range(365):
        for hour in range(24):
            temp_yh[day][hour] = np.random.randint(-4, 30)

    result = hdd_cdd.effective_temps(temp_yh, nr_day_to_av=1)

    expected = (temp_yh[0] + temp_yh[1]) / 2

    expected2 = (expected + temp_yh[2]) / 2

    # positive values
    np.testing.assert_array_equal(result[1], expected)

    np.testing.assert_array_equal(result[2], expected2)

def test_calc_hdd():
    """testing
    """
    t_base = 15 #degree
    temp_yh = np.zeros((365, 24))
    
    for day in range(365):
        for hour in range(24):
            temp_yh[day][hour] = np.random.randint(-4, 30)
    
    # Test with hourly input data
    result = hdd_cdd.calc_hdd(t_base, temp_yh, nr_day_to_av=0, crit_temp_min_max=False)

    temp_yh[temp_yh > t_base] = t_base

    expected = np.sum(t_base - temp_yh) / 24

    # positive values
    assert round(np.sum(result), 3) == round(expected, 3)

def test_calc_cdd():
    """testing
    """
    t_base = 15 #degree
    temp_yh = np.zeros((365, 24))

    for day in range(365):
        for hour in range(24):
            temp_yh[day][hour] = np.random.randint(-4, 30)

    result = hdd_cdd.calc_cdd(t_base, temp_yh, nr_day_to_av=0)

    temp_yh[temp_yh < t_base] = t_base

    expected = np.sum(temp_yh - t_base) / 24

    # positive values
    assert round(np.sum(result), 3) == round(expected, 3)

def test_get_hdd_country():
    """testing
    """

    base_yr = 2015
    curr_yr = 2020

    weather_stations = {
        "weater_station_A": {
            'latitude': 55.8695,
            'longitude': -4.4}}

    regions = ['reg_A', 'reg_B']

    temp_data = {
        "weater_station_A": np.zeros((365, 24)) + 12}

    base_temp_diff_params = {}
    base_temp_diff_params['sig_midpoint'] = 0
    base_temp_diff_params['sig_steepness'] = 1
    base_temp_diff_params['yr_until_changed'] = 2020

    reg_coord = {
        "reg_A": {
            'latitude': 59.02999742,
            'longitude': -3.4},
        "reg_B": {
            'latitude': 57.02999742,
            'longitude': -4.4}}

    t_base_heating_base_yr = 15.5

    result = hdd_cdd.get_hdd_country(
        t_base_heating_base_yr,
        regions,
        temp_data,
        reg_coord,
        weather_stations)

    expected = {
        "reg_A": (15.5 - 12.0) * 8760 / 24,
        "reg_B": (15.5 - 12.0) * 8760 / 24}

    assert result['reg_A'] == expected['reg_A']
    assert result['reg_B'] == expected['reg_B']

def test_get_cdd_country():
    """testing
    """
    weather_stations = {
        "weater_station_A": {
            'latitude': 55.8695,
            'longitude': -4.4}}

    regions = ['reg_A', 'reg_B']

    temp_data = {
        "weater_station_A": np.zeros((365, 24)) + 20}

    base_temp_diff_params = {}
    base_temp_diff_params['sig_midpoint'] = 0
    base_temp_diff_params['sig_steepness'] = 1
    base_temp_diff_params['yr_until_changed'] = 2020

    reg_coord = {
        "reg_A": {
            'latitude': 59.02999742,
            'longitude': -3.4},
        "reg_B": {
            'latitude': 57.02999742,
            'longitude': -4.4}}

    t_base_heating_base_yr = 15.5

    result = hdd_cdd.get_cdd_country(
        t_base_heating_base_yr,
        regions,
        temp_data,
        reg_coord,
        weather_stations)

    expected = {
        "reg_A": (20 - 15.5) * 8760 / 24,
        "reg_B": (20 - 15.5) * 8760 / 24}

    assert result['reg_A'] == expected['reg_A']
    assert result['reg_B'] == expected['reg_B']

def test_calc_reg_hdd():
    """testing
    """
    t_base_heating = 15.5

    temperatures = np.zeros((365, 24)) + 12

    model_yeardays = range(365)

    result_hdd_d, result_shape = hdd_cdd.calc_reg_hdd(temperatures, t_base_heating, model_yeardays)

    assert np.sum(result_hdd_d) == (15.5 - 12) * 8760 / 24
    assert round(np.sum(result_shape), 3) == round(1.0, 3)

def test_calc_reg_cdd():
    """testing
    """
    t_base_heating = 15.5

    temperatures = np.zeros((365, 24)) + 20

    model_yeardays = range(365)

    result_hdd_d, result_shape = hdd_cdd.calc_reg_cdd(temperatures, t_base_heating, model_yeardays)

    assert np.sum(result_hdd_d) == (20 - 15.5) * 8760 / 24
    assert round(np.sum(result_shape), 3) == round(1.0, 3)



"""Function to generate plots based on simulation results stored in a folder
"""
import os
import sys

from energy_demand.basic import basic_functions
from energy_demand.plotting import figs_p2
from energy_demand.plotting import fig_spatial_distribution_of_peak
from energy_demand.plotting import fig_p2_spatial_weather_map
from energy_demand.plotting import fig_p2_total_demand_national_scenarios

def paper_II_plots(
        #/soge-home/staff/cenv0553/ED
        #scenario_path_all_stations_single_weather_yr="C:/Users/cenv0553/ed/results/_multiple_TEST_two_scenarios",
        #scenario_path_all_stations_single_weather_yr="C:/Users/cenv0553/ed/results/_Fig2_multiple_all_yrs_all_stations",
        ed_path='C:/Users/cenv0553/ED',
        scenario_path_single_stations='C:/Users/cenv0553/ED/results/_Fig2_multiple_2015_weather_stations',
        scenario_path_all_stations_single_weather_yr="C:/Users/cenv0553/ED/results/_Fig2_multiple_all_yrs_all_stations",
        path_shapefile_input="C:/Users/cenv0553/ED/data/region_definitions/lad_2016_uk_simplified.shp"
    ):
    """Iterate the folders with scenario
    runs and generate PDF results of individual
    simulation runs

    """
    # Chose which plots should be generated
    plot_crit_dict = {

        # Scenario plots
        'scenario_plots': False, #True,

        #Plot weather map
        'weather_map': False, #True,

        # Spatio temporal validation
        'spatio_temporal_validation': False, #True,

        # Spatial map of distribution of peak and contribution in terms of overall variability
        "plot_spatial_distribution_of_peak": True,

        # Needs to have weatehryr_stationid files in folder
        "plot_national_regional_hdd_disaggregation": False,

        # Plot weather variability for every weather station
        # of demand generated with regional HDD calculation
        "plot_weather_day_year": False,

        # Plot spatial distribution of differences in
        # variability of demand by weather variability
        # normed by population
        "plot_spatial_weather_var_peak": False,

        # Plot scenario years sorted with weather variability
        "plot_scenarios_sorted": True,
        }


    ####################################################################
    # Plot scenarios 
    ####################################################################
    if plot_crit_dict['scenario_plots']:
        sim_yrs = [2015, 2020, 2030, 2040, 2050]
        sim_yrs = [2015, 2050]

        scenario_result_paths = [
            "C:/Users/cenv0553/ed/results/_scenario_low_elec",
            "C:/Users/cenv0553/ed/results/_scenario_high_elec"]
        
        scenario_result_paths = ["C:/Users/cenv0553/ed/results/_Fig2_multiple_all_yrs_all_stations"]
        
        # Create result folder
        path_out_plots = os.path.join("C:/Users/cenv0553/ed/results/_results_nationalFIGS")
        basic_functions.del_previous_setup(path_out_plots)
        basic_functions.create_folder(path_out_plots)
        
        fig_p2_total_demand_national_scenarios.total_demand_national_scenarios(
            scenario_result_paths,
            sim_yrs=sim_yrs,
            fueltype_str='electricity',
            path_out_plots=os.path.join(path_out_plots, "PDF_national_scenarios.pdf"))

    ####################################################################
    # Plot weather station availability map
    ####################################################################
    if plot_crit_dict['weather_map']:

        # Wheather station distribution map
        path_to_weather_data = os.path.join(ed_path, "data/_raw_data/A-temperature_data/cleaned_weather_stations_data")
        folder_path_weater_stations = os.path.join(ed_path, "data/_raw_data/A-temperature_data/cleaned_weather_stations.csv")
        fig_path = os.path.join(ed_path, scenario_path_single_stations, "weather_station_maps.pdf")
        fig_p2_spatial_weather_map.run(
            path_to_weather_data,
            folder_path_weater_stations,
            path_shapefile=path_shapefile_input,
            fig_path=fig_path)
        
        print("... plotted weather station distribution map")

    ####################################################################
    # Plot regional vs national spatio-temporal validation
    ####################################################################
    if plot_crit_dict['spatio_temporal_validation']:

        path_regional_calculations = os.path.join(scenario_path_single_stations, "_regional_calculations")
        path_non_regional_elec_2015 = os.path.abspath(os.path.join(path_regional_calculations, '..', "_all_stations_wy_2015"))
        path_rolling_elec_demand = os.path.join(ed_path, "energy_demand", "energy_demand", "config_data", '01-validation_datasets', '01_national_elec_2015', 'elec_demand_2015.csv')
        path_temporal_elec_validation = os.path.join(ed_path, "energy_demand", "energy_demand", "config_data", '01-validation_datasets', '02_subnational_elec', 'data_2015_elec_domestic.csv')
        path_temporal_gas_validation = os.path.join(ed_path, "energy_demand", "energy_demand", "config_data", '01-validation_datasets', '03_subnational_gas', 'data_2015_gas_domestic.csv')

        path_results = os.path.abspath(os.path.join(path_regional_calculations, '..'))
        path_out_plots = os.path.join(path_results, '_results_PDF_figs')
        basic_functions.del_previous_setup(path_out_plots)
        basic_functions.create_folder(path_out_plots)

        # Plot figure national an regional validation comparison
        figs_p2.plot_fig_spatio_temporal_validation(
            path_regional_calculations=path_regional_calculations,
            path_rolling_elec_demand=path_rolling_elec_demand,
            path_temporal_elec_validation=path_temporal_elec_validation,
            path_temporal_gas_validation=path_temporal_gas_validation,
            path_non_regional_elec_2015=path_non_regional_elec_2015,
            path_out_plots=path_out_plots)

        print("... plotted spatio-temporal validation")

    ####################################################################
    # Plot spatial distribution of variability depending on weather year
    # Plot the variability of the contribution of regional peak demand
    # to national peak demand
    ####################################################################
    if plot_crit_dict['plot_spatial_distribution_of_peak']:

        # Create result folder
        path_out_plots = os.path.join(scenario_path_all_stations_single_weather_yr, '_results_PDF_figs')
        basic_functions.del_previous_setup(path_out_plots)
        basic_functions.create_folder(path_out_plots)

        # Get all folders with scenario run results (name of folder is scenario)
        scenarios = os.listdir(scenario_path_all_stations_single_weather_yr)

        scenario_names_ignored = [
            '__results_multiple_scenarios',
            '_FigII_non_regional_2015',
            '_results_PDF_figs']

        only_scenarios = []
        for scenario in scenarios:
            if scenario in scenario_names_ignored:
                pass
            else:
                only_scenarios.append(scenario)

        # Select simulation years
        sim_yrs = [2015, 2050]
        unit = 'kW' #'GW'

        # Plot standard deviation of demand per person in peak hour
        fig_spatial_distribution_of_peak.run_fig_spatial_distribution_of_peak(
            only_scenarios,
            scenario_path_all_stations_single_weather_yr,
            path_shapefile_input,
            sim_yrs=sim_yrs,
            field_to_plot="std_deviation_df_abs_demand_in_peak_h_pp",
            unit=unit,
            fig_path=path_out_plots)

        # Plot standard deviation of percentage of national peak
        fig_spatial_distribution_of_peak.run_fig_spatial_distribution_of_peak(
            only_scenarios,
            scenario_path_all_stations_single_weather_yr,
            path_shapefile_input,
            sim_yrs=sim_yrs,
            field_to_plot='std_deviation_p_demand_peak_h',
            unit='percent',
            fig_path=path_out_plots)

        # Plot standard deviation of absolute demand of national peak
        fig_spatial_distribution_of_peak.run_fig_spatial_distribution_of_peak(
            only_scenarios,
            scenario_path_all_stations_single_weather_yr,
            path_shapefile_input,
            sim_yrs=sim_yrs,
            field_to_plot='std_deviation_abs_demand_peak_h',
            unit=unit,
            fig_path=path_out_plots)

        # Plot maximum peak hour - mean peak hour demand
        fig_spatial_distribution_of_peak.run_fig_spatial_distribution_of_peak(
            only_scenarios,
            scenario_path_all_stations_single_weather_yr,
            path_shapefile_input,
            sim_yrs=sim_yrs,
            field_to_plot='diff_peak_h_minus_mean',
            unit='GW',
            fig_path=path_out_plots)

        print("... plotted variability maps")

    
    print("Finihsed")
    raise Exception("Finished")

    ####################################################################
    # left overs
    ####################################################################
    for scenario in only_scenarios:

        # -----------
        # Other plots
        # -----------
        figs_p2.main(
            os.path.join(scenario_path_all_stations_single_weather_yr, scenario),
            path_shapefile_input,
            path_out_plots,
            plot_crit_dict)

    # ####################################################################
    # Plot demand over time and peak over time (for modassar paper)
    # ####################################################################
    '''if plot_crit_dict['plot_weather_day_year']:

        # --------------------------------------------
        # Plot peak demand and total demand per fueltype
        # --------------------------------------------
        for fueltype_str in data['lookups']['fueltypes'].keys():

            fig_total_demand_peak.run(
                data_input=weather_yr_container['tot_fueltype_yh'],
                fueltype_str=fueltype_str,
                fig_name=os.path.join(
                    path_out_plots, "tot_{}_h.pdf".format(fueltype_str)))

        # plot over period of time across all weather scenario
        fig_weather_variability_priod.run(
            data_input=weather_yr_container['tot_fueltype_yh'],
            fueltype_str='electricity',
            simulation_yr_to_plot=2015, # Simulation year to plot
            period_h=list(range(200,500)), #period to plot
            fig_name=os.path.join(
                path_out_plots, "weather_var_period.pdf"))'''

    return

if __name__ == '__main__':
    # Map command line arguments to function arguments.
    paper_II_plots(*sys.argv[1:])



"""All fuel shares of the base year for the
different technologies are defined in this file.
"""
from energy_demand.initalisations import helpers

def assign_by_fuel_tech_p(
        enduses,
        sectors,
        fueltypes,
        fueltypes_nr
    ):
    """Assigning fuel share per enduse for different technologies
    for the base year.

    Arguments
    ----------
    enduses : dict
        Enduses
    sectors : dict
        Sectors per submodel
    fueltypes : dict
        Fueltypes lookup
    fueltypes_nr : int
        Number of fueltypes

    Returns
    -------
    fuel_tech_p_by : dict
        Residential fuel share percentages

    Note
    ----
    - In an enduse, either all fueltypes with assigned fuelsneed to be
      assigned with technologies or none. No mixing possible

    - Technologies can be defined for the following fueltypes:
        'solid_fuel': 0,
        'gas': 1,
        'electricity': 2,
        'oil': 3,
        'biomass': 4,
        'hydrogen': 5,
        'heat': 6

    -   Not defined fueltypes will be assigned placholder technologies
    """
    fuel_tech_p_by = {}

    _fuel_tech_p_by = helpers.init_fuel_tech_p_by(
        enduses['residential'], fueltypes_nr)
    fuel_tech_p_by.update(_fuel_tech_p_by)

    _fuel_tech_p_by = helpers.init_fuel_tech_p_by(
        enduses['service'], fueltypes_nr)
    fuel_tech_p_by.update(_fuel_tech_p_by)

    _fuel_tech_p_by = helpers.init_fuel_tech_p_by(
        enduses['industry'], fueltypes_nr)
    fuel_tech_p_by.update(_fuel_tech_p_by)

    # ====================
    # Residential Submodel
    # ====================

    # ---------------
    # rs_lighting
    # Calculated on the basis of ECUK Table 3.08
    # ---------------
    fuel_tech_p_by['rs_lighting'][fueltypes['electricity']] = {
        'standard_lighting_bulb': 0.04,
        'halogen': 0.56,
        'fluorescent_strip_lighting': 0.07,
        'energy_saving_lighting_bulb': 0.32,
        'LED': 0.01}

    # ---------------
    # rs_cold
    # Calculated on the basis of ECUK Table 3.08
    # ---------------
    fuel_tech_p_by['rs_cold'][fueltypes['electricity']] = {
        'chest_freezer': 0.087,
        'fridge_freezer': 0.588,
        'refrigerator': 0.143,
        'upright_freezer': 0.182}

    # ---------------
    # rs_cooking
    # Calculated on the basis of ECUK Table 3.08
    # Calculated on the assumption that 5 to 10%
    # of all households have induction hobs (https://productspy.co.uk/are-induction-hobs-safe/ (5-10%))
    # ---------------
    fuel_tech_p_by['rs_cooking'][fueltypes['electricity']] = {
        'hob_electricity': 0.95,
        'hob_induction_electricity': 0.05}
    fuel_tech_p_by['rs_cooking'][fueltypes['gas']] = {
        'hob_gas': 1.0}
    fuel_tech_p_by['rs_cooking'][fueltypes['hydrogen']] = {
        'hob_hydrogen': 1.0}
    fuel_tech_p_by['rs_cooking'][fueltypes['biomass']] = {
        'hob_biomass': 1.0}

    # ---------------
    # rs_wet
    # calculated on the basis of EUCK Table 3.08
    # ---------------
    fuel_tech_p_by['rs_wet'][fueltypes['electricity']] = {
        'washing_machine': 0.305,
        'washer_dryer': 0.157,
        'dishwasher': 0.220,
        'tumble_dryer': 0.318}

    # ---------------
    # rs_space_heating
    #
    # According to the DCLG (2014) English Housing Survey. Energy Report. doi: 10.1017/CBO9781107415324.004.
    # Annex Table 3.1, the following number of electric heating technologies can be found in the UK:
    #
    # storage heaters             5.5   % of all houses
    # electric room heaters	      2.0   % of all houses
    # electric central heating	 0.65   % of all houses
    #
    # As heat pumps were not accounted for, they are taken from OFGEM (2015),
    # which states that there are about 0.1m heat pumps of about in total 27m
    # households in the UK. This corresponds to about 0.4 %. (see also Hannon 2015).
    # According to Hannon (2015), heat pumps account only for a tiny fraction of the UK.
    # heat supply for buildings (approximately 0.2%). This percentage is substract from
    # the storage heaters.
    #
    # storage heaters                   5.1   % of all houses --> ~ 62%     (100.0 / 8.15) * 5.1
    # secondary_heater_electricity
    #   electric room heaters	        2.0   % of all houses --> ~ 25%     (100.0 / 8.15) * 2.0
    #   electric central heating	    0.65  % of all houses --> ~ 8%     (100.0 / 8.15) * 0.65
    # heat pumps                        0.4   % of all houses --> ~ 0.5%     (100.0 / 8.15) * 0.4
    #
    # OFGEM (2015); Insights paper on households with electric and other non-gas heating,
    # (December), 1–84.
    #
    # Hannon, M. J. (2015). Raising the temperature of the UK heat pump market:
    # Learning lessons from Finland. Energy Policy, 85, 369–375.
    # https://doi.org/10.1016/j.enpol.2015.06.016
    # ---------------
    fuel_tech_p_by['rs_space_heating'][fueltypes['solid_fuel']] = {
        'boiler_solid_fuel': 1.0}

    fuel_tech_p_by['rs_space_heating'][fueltypes['oil']] = {
        'boiler_condensing_oil': 0.6,
        'boiler_oil': 0.4}

    # ---
    # According to table 3.19, 59.7% (43.5% + 14.3%) have some form of condensing boiler.

    # Todays share of district heating is about 2% of UK non-industraiyl demand
    # http://fes.nationalgrid.com/media/1215/160712-national-grid-dh-summary-report.pdf
    # ---
    fuel_tech_p_by['rs_space_heating'][fueltypes['gas']] = {
        'boiler_condensing_gas': 0.60,
        'boiler_gas': 0.37,
        'district_heating_CHP_gas': 0.03}

    fuel_tech_p_by['rs_space_heating'][fueltypes['electricity']] = {
        'district_heating_electricity' : 0,
        'storage_heater_electricity': 0.62,
        'secondary_heater_electricity':0.33,
        'heat_pumps_electricity': 0.05}

    fuel_tech_p_by['rs_space_heating'][fueltypes['biomass']] = {
        'boiler_biomass': 1.0,
        'district_heating_biomass': 0.0}

    fuel_tech_p_by['rs_space_heating'][fueltypes['hydrogen']] = {
        'fuel_cell_hydrogen': 0,
        'district_heating_fuel_cell': 0,
        'boiler_hydrogen': 1.0,
        'heat_pumps_hydrogen': 0.0}

    # -------------
    # Residential water heating
    # -------------
    fuel_tech_p_by['rs_water_heating'][fueltypes['gas']] = {
        'boiler_condensing_gas': 0.60,
        'boiler_gas': 0.37,
        'district_heating_CHP_gas': 0.03}

    fuel_tech_p_by['rs_water_heating'][fueltypes['electricity']] = {
        'storage_heater_electricity': 0.62,
        'secondary_heater_electricity':0.33,
        'heat_pumps_electricity': 0.05}

    fuel_tech_p_by['rs_water_heating'][fueltypes['biomass']] = {
        'boiler_biomass': 1.0,
        'district_heating_biomass': 0.0}

    fuel_tech_p_by['rs_water_heating'][fueltypes['hydrogen']] = {
        'boiler_hydrogen': 1.0}

    fuel_tech_p_by['rs_water_heating'][fueltypes['oil']] = {
        'boiler_oil': 1.0}

    fuel_tech_p_by['rs_water_heating'][fueltypes['solid_fuel']] = {
        'boiler_solid_fuel': 1.0}

    # ===================
    # Service subModel
    # ===================

    # ss_lighting Simplified based on Table 5.09 (Office lighting)
    fuel_tech_p_by['ss_lighting'][fueltypes['electricity']] = {
        'halogen': 0.45,
        'fluorescent_strip_lighting': 0.07,
        'energy_saving_lighting_bulb': 0.47, #All different lighting next to halogen are summarised here ("non-halogen lighting")
        'LED': 0.01}

    # ----------------
    # Service space heating (ss_space_heating)
    #  For ss_space heating the load profile is the same for all technologies
    # ----------------
    fuel_tech_p_by['ss_space_heating'][fueltypes['solid_fuel']] = {
        'boiler_solid_fuel': 1.0}

    fuel_tech_p_by['ss_space_heating'][fueltypes['gas']] = {
        'district_heating_CHP_gas': 0.02,
        'boiler_condensing_gas': 0.6,
        'boiler_gas': 0.38}

    fuel_tech_p_by['ss_space_heating'][fueltypes['electricity']] = {
        'district_heating_electricity' : 0,
        'secondary_heater_electricity': 0.95,
        'heat_pumps_electricity': 0.05}

    fuel_tech_p_by['ss_space_heating'][fueltypes['oil']] = {
        'boiler_condensing_oil': 0.6,
        'boiler_oil': 0.4}

    fuel_tech_p_by['ss_space_heating'][fueltypes['biomass']] = {
        'boiler_biomass': 1.0}

    fuel_tech_p_by['ss_space_heating'][fueltypes['hydrogen']] = {
        'fuel_cell_hydrogen': 0,
        'boiler_hydrogen': 1.0,
        'heat_pumps_hydrogen': 0.0,
        'district_heating_fuel_cell': 0.0}

    # -------------
    # Service water heating
    # -------------
    fuel_tech_p_by['ss_water_heating'][fueltypes['gas']] = {
        'boiler_condensing_gas': 0.60,
        'boiler_gas': 0.37,
        'district_heating_CHP_gas': 0.03}

    fuel_tech_p_by['ss_water_heating'][fueltypes['electricity']] = {
        'storage_heater_electricity': 0.62,
        'secondary_heater_electricity':0.33,
        'heat_pumps_electricity': 0.05}

    fuel_tech_p_by['ss_water_heating'][fueltypes['biomass']] = {
        'boiler_biomass': 1.0,
        'district_heating_biomass': 0.0}

    fuel_tech_p_by['ss_water_heating'][fueltypes['hydrogen']] = {
        'boiler_hydrogen': 1.0}

    fuel_tech_p_by['ss_water_heating'][fueltypes['oil']] = {
        'boiler_oil': 1.0}

    fuel_tech_p_by['ss_water_heating'][fueltypes['solid_fuel']] = {
        'boiler_solid_fuel': 1.0}

    # ------------------------------
    # Cooling
    # ECUK Table 5.09
    # ------------------------------
    fuel_tech_p_by['ss_cooling_humidification'][fueltypes['electricity']] = {
        'central_air_conditioner_electricity': 0.64,
        'decentral_air_conditioner_electricity': 0.36}
    fuel_tech_p_by['ss_cooling_humidification'][fueltypes['gas']] = {
        'central_air_conditioner_gas': 0.64,
        'decentral_air_conditioner_gas': 0.36}
    fuel_tech_p_by['ss_cooling_humidification'][fueltypes['oil']] = {
        'central_air_conditioner_oil': 0.64,
        'decentral_air_conditioner_oil': 0.36}

    # Helper: Transfer all defined shares for every enduse to every sector
    fuel_tech_p_by = helpers.copy_fractions_all_sectors(
        fuel_tech_p_by,
        sectors['service'],
        affected_enduses=enduses['service'])

    # ===================
    # Industry subModel  - Fuel shares of technologies in enduse
    # ===================

    # ----------------
    # Industrial space heating (is_space_heating)
    # ----------------
    fuel_tech_p_by['is_space_heating'][fueltypes['solid_fuel']] = {
        'boiler_solid_fuel': 1.0}

    fuel_tech_p_by['is_space_heating'][fueltypes['gas']] = {
        'district_heating_CHP_gas': 0.02,
        'boiler_condensing_gas': 0.6,
        'boiler_gas': 0.38}

    fuel_tech_p_by['is_space_heating'][fueltypes['electricity']] = {
        'district_heating_electricity' : 0,
        'secondary_heater_electricity': 0.95,
        'heat_pumps_electricity': 0.05,
        'storage_heater_electricity': 0}

    fuel_tech_p_by['is_space_heating'][fueltypes['oil']] = {
        'boiler_condensing_oil': 0.6,
        'boiler_oil': 0.4}

    fuel_tech_p_by['is_space_heating'][fueltypes['biomass']] = {
        'boiler_biomass': 1.0}

    fuel_tech_p_by['is_space_heating'][fueltypes['hydrogen']] = {
        'fuel_cell_hydrogen': 0,
        'boiler_hydrogen': 1.0,
        'heat_pumps_hydrogen': 0.0,
        'district_heating_fuel_cell': 0.0}

    # Helper: Transfer all defined shares for every enduse to every sector
    fuel_tech_p_by = helpers.copy_fractions_all_sectors(
        fuel_tech_p_by,
        sectors=sectors['industry'],
        affected_enduses=enduses['industry'])

    # ----------------
    # Industrial High temporal processes (is_high_temp_process)
    # ----------------
    # Todays share is about: 17% electric furnace, 82% basic oxygen (Key Statistics 2016, appea, EnergyQuest)
    #-- basic_metals (sector)
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['solid_fuel']] = {
        'basic_oxygen_furnace': 1.0}
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['electricity']] = {
        'electric_arc_furnace': 1.0}
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['gas']] = {
        'SNG_furnace': 1.0}
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['biomass']] = {
        'biomass_furnace': 1.0}
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['hydrogen']] = {
        'hydrogen_furnace': 1.0}
    fuel_tech_p_by['is_high_temp_process']['basic_metals'][fueltypes['oil']] = {
        'oil_furnace': 1.0}

    #-- non_metallic_mineral_products
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['solid_fuel']] = {
        'dry_kiln_coal': 0.9,
        'wet_kiln_coal': 0.1}
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['oil']] = {
        'dry_kiln_oil': 0.9,
        'wet_kiln_oil': 0.1}
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['gas']] = {
        'dry_kiln_gas': 0.9,
        'wet_kiln_gas': 0.1}
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['electricity']] = {
        'dry_kiln_electricity': 0.9,
        'wet_kiln_electricity': 0.1}
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['biomass']] = {
        'dry_kiln_biomass': 1.0}
    fuel_tech_p_by['is_high_temp_process']['non_metallic_mineral_products'][fueltypes['hydrogen']] = {
        'dry_kiln_hydrogen': 1.0}

    return dict(fuel_tech_p_by)



"""This file contains all calculations related
to spatial explicit calculations of technology/innovation
penetration."""
import logging
from collections import defaultdict
import numpy as np

def spatial_diffusion_values(
        regions,
        real_values,
        speed_con_max,
        low_congruence_crit,
        p_outlier
    ):
    """Generate spatial diffusion values from real data

    Arguments
    ---------
    regions : dict
        Regions
    p_outlier : float (percentage)
        Percentage of outliers which are capped at both
        ends of the value spectrum of the real data

    Returns
    -------
    diffusion_values : dict
        Spatial diffusion values based on speed assumptions

    Example
    -------
    This function calculates the values which already
    incorporate different speeds in diffusion.

    For example based on real values (e.g. population density)
    congruence values are calculated. Then, the congruence
    values are linked to diffusion speed differentes.
    """
    diffusion_values = {}

    # Diffusion speed assumptions
    speed_con_min = 1               # Speed at val_con == 0
    speed_con_max = speed_con_max   # Speed at con_val == 1

    if speed_con_max == 1:

        # No regional difference
        for region in regions:
            diffusion_values[region] = 1 #100% congruence
    else:
        # ----------------
        # plot real values to check for outliers
        # ----------------

        # Select number of outliers to remove lower and higher extremes
        nr_of_outliers = int(100 / len(regions) * p_outlier)

        sorted_vals = list(real_values.values())
        sorted_vals.sort()

        # Get value of largest outlier
        treshold_upper_real_value = sorted_vals[-nr_of_outliers]
        treshold_lower_real_value = sorted_vals[nr_of_outliers]

        for reg, val in real_values.items():
            if val > treshold_upper_real_value:
                real_values[reg] = treshold_upper_real_value
            if val < treshold_lower_real_value:
                real_values[reg] = treshold_lower_real_value

        # ---------------------------------
        # Congruence calculations
        # ----------------------------------
        # Max congruence value
        con_max = max(real_values.values())

        for region in regions:

            # Multiply speed of diffusion of concept with concept congruence value
            try:
                real_value = real_values[region]
            except KeyError:
                real_value = np.average(real_values.values())
                logging.warning("Set average real data for region %s", region)

            # Calculate congruence value
            congruence_value = real_value / con_max

            # If the assignement is thoe other way round (lowest value has highest congruence value)
            if low_congruence_crit:
                congruence_value = 1 - congruence_value
            else:
                pass

            # Calculate diffusion value
            lower_concept_val = (1 - congruence_value) * speed_con_min
            higher_concept_val = congruence_value * speed_con_max

            diffusion_values[region] = lower_concept_val + higher_concept_val

    return diffusion_values

def calc_diffusion_f(regions, f_reg, spatial_diff_values, fuels):
    """From spatial diffusion values calculate diffusion
    factor for every region (which needs to sum up to one
    across all regions) and end use. With help of these calculation diffusion
    factors, a spatial explicit diffusion of innovations can be
    implemented.

    Arguments
    ----------
    regions : dict
        Regions
    f_reg : dict
        Regional not weighted diffusion factors
    spatial_diff_values : dict
        Spatial diffusion index values
    fuels : array
        Fuels per enduse or fuel per sector and enduse

    Example
    -------
    If the national assumption of a technology diffusion
    of 50% is defined (e.g. 50% of service are heat pumps), this
    percentage can be changed per region, i.e. in some regions
    with higher diffusion factors, a larger percentage adopt
    the technology on the expense of other regions, where a
    lower percentage adopt this technology. In sum however,
    for all regions, the total service still sums up to 50%.

    Note
    -----
    The total sum can be higher than 1 in case of high values.
    Therefore the factors need to be capped. TODO MORE INFO
    """
    # Calculate fraction of energy demand of every region of total demand
    reg_enduse_p = defaultdict(dict)
    fuels_enduse = {}

    for fuel_submodel in fuels:

        # -----------------------------------
        # Sum fuel across sectors
        # -----------------------------------
        fuel_submodel_new = defaultdict(dict)
        for region, entries in fuel_submodel.items():
            enduses = entries.keys()
            try:
                for enduse in entries:
                    for sector in entries[enduse]:
                        fuel_submodel_new[region][enduse] = 0

                for enduse in entries:
                    for sector in entries[enduse]:
                        fuel_submodel_new[region][enduse] += np.sum(entries[enduse][sector])

                fuel_submodel = fuel_submodel_new
            except IndexError:
                enduses = entries.keys()
                break

        # --------------------
        # Calculate fraction of fuel for each region
        # --------------------
        for enduse in enduses:
            fuels_enduse[enduse] = 0

            # Total uk fuel of enduse
            tot_enduse_uk = 0
            for region in regions:
                tot_enduse_uk += np.sum(fuel_submodel[region][enduse])

            # Calculate regional % of enduse
            for region in regions:
                reg_enduse_p[enduse][region] = np.sum(fuel_submodel[region][enduse]) / tot_enduse_uk
                fuels_enduse[enduse] += np.sum(fuel_submodel[region][enduse])

        # ----------
        # Norm spatial factor (f_reg_norm) with population (does not sum upt to 1.p Eq. 7 Appendix)
        # ----------
        f_reg_norm = {}

        for enduse, regions_fuel_p in reg_enduse_p.items():

            # Sum across all regs (factor * fuel_p)
            sum_p_f_all_regs = 0
            for region in regions:
                sum_p_f_all_regs += f_reg[region] * regions_fuel_p[region]

            f_reg_norm[enduse] = {}
            for region, fuel_p in regions_fuel_p.items():
                f_reg_norm[enduse][region] = f_reg[region] / sum_p_f_all_regs

        # ----------
        # Norm which sums up to 1 (f_reg_norm_abs) (e.g. distriubte 200 units across space)
        # ----------
        f_reg_norm_abs = {}
        for enduse, regions_fuel_p in reg_enduse_p.items():
            f_reg_norm_abs[enduse] = {}
            for region, fuel_p in regions_fuel_p.items():
                f_reg_norm_abs[enduse][region] = fuel_p * spatial_diff_values[region]

    #-----------
    # Normalize f_reg_norm_abs
    #-----------
    for enduse in f_reg_norm_abs:
        sum_enduse = sum(f_reg_norm_abs[enduse].values())
        for region in f_reg_norm_abs[enduse]:
            f_reg_norm_abs[enduse][region] = f_reg_norm_abs[enduse][region] / sum_enduse

    # Testing
    for enduse in f_reg_norm_abs:
        np.testing.assert_almost_equal(
            sum(f_reg_norm_abs[enduse].values()),
            1,
            decimal=2)

    return f_reg_norm_abs, f_reg_norm

def calc_regional_services(
        enduse,
        uk_techs_service_p,
        regions,
        spatial_factors,
        fuel_disaggregated,
        techs_affected_spatial_f,
        capping_val=1
    ):
    """Calculate regional specific end year service shares
    of technologies (rs_reg_enduse_tech_p_ey)

    Arguments
    ---------
    uk_techs_service_p : dict
        Service shares per technology for future year
    regions : dict
        Regions
    spatial_factors : dict
        Spatial factor per enduse and region
    fuel_disaggregated : dict
        Fuels per region
    techs_affected_spatial_f : list
        List with technologies where spatial diffusion is affected
    capping_val : float
        Maximum service share (1.0). This is needed in case
        of spatial explicit diffusion modelling where the diffusion
        speed is very large and thus would lead to areas with
        largher shares than 1

    Returns
    -------
    rs_reg_enduse_tech_p_ey : dict
        Regional specific model end year service shares of techs

    Modelling steps
    -----
    A.) Calculation national end use service to reduce
        (e.g. 50% heat pumps for all regions) (uk_tech_service_ey_p)

    B.) Distribute this service according to spatial index for
        techs where the spatial explicit diffusion applies (techs_affected_spatial_f).
        Otherwise disaggregated according to fuel

    C.) Convert regional service reduction to ey % in region
    """
    reg_enduse_tech_p_ey = defaultdict(dict)

    # ------------------------------------
    # Calculate national total enduse fuel and service
    # ------------------------------------
    uk_enduse_fuel = 0
    for region in regions:
        reg_enduse_tech_p_ey[region] = {}
        uk_enduse_fuel += np.sum(fuel_disaggregated[region][enduse])

    # ----
    # Service of enduse for all regions
    # ----
    for region in regions:

        # Disaggregation factor
        f_fuel_disagg = np.sum(fuel_disaggregated[region][enduse]) / uk_enduse_fuel

        # Calculate fraction of regional service
        for tech, uk_tech_service_ey_p in uk_techs_service_p.items():

            global_tech_service_ey_p = uk_tech_service_ey_p

            # ---------------------------------------------
            # B.) Calculate regional service for technology
            # ---------------------------------------------
            if tech in techs_affected_spatial_f:
                # Use spatial factors
                reg_service_tech = global_tech_service_ey_p * spatial_factors[enduse][region]
            else:
                # If not specified, use fuel disaggregation for enduse factor
                reg_service_tech = global_tech_service_ey_p #* f_fuel_disagg

            reg_enduse_tech_p_ey[region][tech] = reg_service_tech

        # ---------------------------------------------
        # C.) Calculate regional fraction
        # ---------------------------------------------
        for tech, service_tech in reg_enduse_tech_p_ey[region].items():

            # ----------------------------------
            # Capping value in case larger than 1.0
            # ----------------------------------
            service_share = service_tech

            if service_share > capping_val:
                reg_enduse_tech_p_ey[region][tech] = capping_val
                logging.info("Maximum value is capped: {} {} {}".format(
                    region, service_share, tech))
            else:
                reg_enduse_tech_p_ey[region][tech] = service_share

    return dict(reg_enduse_tech_p_ey)

def calc_spatially_diffusion_factors(
        regions,
        fuel_disagg,
        real_values,
        low_congruence_crit,
        speed_con_max,
        p_outlier
    ):
    """
    Calculate spatial diffusion values

    Arguments
    ---------
    regions : dict
        Regions
    fuel_disagg : dict
        Disaggregated fuel per region
    real_values : dict
        Real values
    p_outlier : float
        Percentage of min and max outliers are flattened

    Returns
    -------
    f_reg_norm_abs : dict
        Diffusion values with normed population. If no value
        is larger than 1, the total sum of all shares calculated
        for every region is identical to the defined scenario variable.

    spatial_diff_values : dict
        Spatial diffusion values (not normed, only considering differences
        in speed and congruence values)

    Explanation
    ============
    (I)     Load diffusion values
    (II)    Calculate diffusion factors
    (III)   Calculate sigmoid diffusion values for technology
            specific enduse service shares for every region
    """
    # -----
    # I. Diffusion diffusion values
    # -----
    spatial_diff_values = spatial_diffusion_values(
        regions=regions,
        real_values=real_values,
        speed_con_max=speed_con_max,
        low_congruence_crit=low_congruence_crit,
        p_outlier=p_outlier)

    # -----
    # II. Calculation of diffusion factors (Not weighted with demand)
    # -----

    # Not weighted with demand
    max_value_diffusion = max(list(spatial_diff_values.values()))
    f_reg = {}
    for region in regions:
        f_reg[region] = spatial_diff_values[region] / max_value_diffusion

    # Weighted with demand
    f_reg_norm_abs, f_reg_norm = calc_diffusion_f(
        regions,
        f_reg,
        spatial_diff_values,
        [fuel_disagg['residential'], fuel_disagg['service'], fuel_disagg['industry']])

    return f_reg, f_reg_norm, f_reg_norm_abs

'''def spatially_differentiated_modelling(
        regions,
        fuel_disagg,
        rs_share_s_tech_ey_p,
        ss_share_s_tech_ey_p,
        is_share_s_tech_ey_p,
        techs_affected_spatial_f,
        spatial_diffusion_factor,
        spatial_explicit_diffusion=False
    ):
    """
    Regional diffusion shares of technologies is calculated
    based on calcualted spatial diffusion factors

    Arguments
    ---------
    regions : dict
        Regions
    fuel_disagg : dict
        Fuel per region
    rs_share_s_tech_ey_p : dict
        Global technology service shares
    ss_share_s_tech_ey_p : dict
        Global technology service shares
    is_share_s_tech_ey_p : dict
        Global technology service shares
    techs_affected_spatial_f : list
        Technologies which are affected by spatially heterogeneous diffusion
    spatial_diffusion_factor : dict
        Spatial diffusion factor

    Returns
    --------
    XX_reg_share_s_tech_ey_p :
        Technology specific service shares for every region (residential)
        considering differences in diffusion speed.

        If the calculate regional shares are larger than 1.0, the
        diffusion is set to the maximum criteria (`cap_max`).
        This means that if some regions reach the maximum defined value,
        thes cannot futher increase their share. This means that other regions diffuse
        slower and do not reach such high leves (and because the faster regions
        cannot over-compensate, the total sum is not identical).

    Calculate sigmoid diffusion values for technology
    specific enduse service shares for every region
    """
    # Residential spatial explicit modelling
    rs_reg_share_s_tech_ey_p = {}
    for enduse, uk_techs_service_p in rs_share_s_tech_ey_p.items():

        rs_reg_share_s_tech_ey_p[enduse] = calc_regional_services(
            enduse,
            uk_techs_service_p,
            regions,
            spatial_diffusion_factor,
            fuel_disagg['residential'],
            techs_affected_spatial_f)

    ss_reg_share_s_tech_ey_p = {}
    for sector, uk_techs_service_enduses_p in ss_share_s_tech_ey_p.items():
        ss_reg_share_s_tech_ey_p[sector] = {}
        for enduse, uk_techs_service_p in uk_techs_service_enduses_p.items():
            ss_reg_share_s_tech_ey_p[sector][enduse] = calc_regional_services(
                enduse,
                uk_techs_service_p,
                regions,
                spatial_diffusion_factor,
                fuel_disagg['ss_fuel_disagg_sum_all_sectors'],
                techs_affected_spatial_f)

    is_reg_share_s_tech_ey_p = {}
    for sector, uk_techs_service_enduses_p in is_share_s_tech_ey_p.items():
        is_reg_share_s_tech_ey_p[sector] = {}
        for enduse, uk_techs_service_p in uk_techs_service_enduses_p.items():
            is_reg_share_s_tech_ey_p[sector][enduse] = calc_regional_services(
                enduse,
                uk_techs_service_p,
                regions,
                spatial_diffusion_factor,
                fuel_disagg['is_aggr_fuel_sum_all_sectors'],
                techs_affected_spatial_f)

    return rs_reg_share_s_tech_ey_p, ss_reg_share_s_tech_ey_p, is_reg_share_s_tech_ey_p'''

def factor_improvements_single(
        factor_uk,
        regions,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs,
        fuel_regs_enduse
    ):
    """Calculate regional specific end year service shares
    of technologies (rs_reg_enduse_tech_p_ey)

    Arguments
    =========
    factor_uk : float
        Improvement of either an enduse or a variable for the whole UK
    regions : dict
        Regions
    f_reg : dict
        Regional spatial factors not normed with fuel demand
    f_reg_norm : dict
        Regional spatial factors normed with fuel demand (sum is not 1)
    f_reg_norm_abs : dict
        Regional spatial factors normed with fuel demand and normed that sum is 1
    spatial_diff_values : dict
        Spatial diffusion values
    fuel_regs_enduse : dict
        Fuels per region and end use

    Returns
    -------
    rs_reg_enduse_tech_p_ey : dict
        Regional specific model end year service shares of techs

    Modelling steps
    -----
    A.) Calculation national end use service to reduce
        (e.g. 50% heat pumps for all regions) (uk_tech_service_ey_p)

    B.) Distribute this service according to spatial index for
        techs where the spatial explicit diffusion applies (techs_affected_spatial_f).
        Otherwise disaggregated according to fuel

    C.) Convert regional service reduction to ey % in region
    """
    reg_enduse_tech_p_ey = {}

    # Check which factors is to be used
    # if only distribute:               f_reg_norm_abs
    # if max 1:                         f_reg_nrm
    # if not intersted in correct sum:  f_reg
    if fuel_regs_enduse == {}:
        logging.info("spatial_factor: fuel_regs_enduse")
        spatial_factor = f_reg
    else:
        logging.info("spatial_factor: f_reg_norm_abs")
        spatial_factor = f_reg_norm_abs

    # Sum fuel for all regions
    uk_enduse_fuel = sum(fuel_regs_enduse.values())

    test = 0
    for region in regions:

        try:
            test += (reg_enduse_tech_p_ey[region] * np.sum(fuel_regs_enduse[region]))
            logging.info(
                "FUEL FACTOR reg: {}  val: {}, fuel: {}  fuel: {} ".format(
                    region,
                    round(reg_enduse_tech_p_ey[region], 3),
                    round(uk_enduse_fuel, 3),
                    round(np.sum(fuel_regs_enduse[region]), 3)))
        except:
            pass

        reg_enduse_tech_p_ey[region] = factor_uk * spatial_factor[region]

        logging.info("spatial single factor reg: {}  val: {}".format(
            region,
            round(reg_enduse_tech_p_ey[region], 3)))

    # ---------
    # PROBLEM THAT MORE THAN 100 percent could be reached if nt normed
    # ---------
    reg_enduse_tech_p_ey_capped = {}

    # Cap regions which have already reached and are larger than 1.0
    cap_max_crit = 1.0 #100%
    demand_lost = 0

    for region, region_factor in reg_enduse_tech_p_ey.items():
        if region_factor > cap_max_crit:
            logging.warning("INFO: FOR A REGION THE SHARE OF IMPROVEMENTIS LARGER THAN 1.0.")

            # Demand which is lost and capped
            diff_to_cap = region_factor - cap_max_crit
            demand_lost += diff_to_cap * np.sum(fuel_regs_enduse[region])

            reg_enduse_tech_p_ey_capped[region] = cap_max_crit
        else:
            reg_enduse_tech_p_ey_capped[region] = region_factor

    # Replace
    reg_enduse_tech_p_ey = reg_enduse_tech_p_ey_capped

    #logging.warning("FAKTOR UK :" + str(factor_uk))
    #logging.warning("Lost demand: " + str(demand_lost))
    #logging.warning("TESTDUM a " + str(test))
    #logging.warning("TESTDUM b " + str(uk_enduse_fuel * factor_uk))
    return reg_enduse_tech_p_ey

def get_enduse_regs(
        enduse,
        fuels_disagg):
    """
    Get a specific enduse for all regions

    Arguments
    ---------
    enduse : str
        Enduse to sum
    fuels_disagg : list
        Fuels per disaggregated regions

    Returns
    -------
    fuels_enduse : dict
        Fuels of an enduse for all regions
        {'reg': np.array(enduse_fuel)}
    """
    fuels_enduse = {}

    for fuel_submodel in fuels_disagg:
        for reg, enduse_fuels in fuel_submodel.items():
            for enduse_to_match, fuels_regs in enduse_fuels.items():
                if enduse == enduse_to_match:
                    fuels_enduse[reg] = fuels_regs

    return fuels_enduse



"""Basic plotting functions
"""
import math
import numpy as np

from scipy.interpolate import interp1d
from scipy.interpolate import spline

from energy_demand.technologies import diffusion_technologies
#matplotlib.use('Agg') # Used to make it work in linux

def export_legend(legend, filename="legend.png"):
    """Export legend as seperate file
    """
    fig = legend.figure
    fig.canvas.draw()
    bbox = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(filename, dpi="figure", bbox_inches=bbox)

def cm2inch(*tupl):
    """Convert input cm to inches (width, hight)
    """
    inch = 2.54
    if isinstance(tupl[0], tuple):
        return tuple(i/inch for i in tupl[0])
    else:
        return tuple(i/inch for i in tupl)

def simple_smooth(x_list, y_list, num=500, spider=False, interpol_kind='quadratic'):

    nr_x_values = len(x_list)
    min_x_val = min(x_list)
    max_x_val = max(x_list)

    x_values = np.linspace(min_x_val, max_x_val, num=nr_x_values, endpoint=True)

    f2 = interp1d(x_values, y_list, kind=interpol_kind)

    smoothed_data_x = np.linspace(
        min_x_val,
        max_x_val,
        num=num,
        endpoint=True)

    smoothed_data_y = f2(smoothed_data_x)

    return smoothed_data_x, smoothed_data_y

def smooth_data(
        x_list,
        y_list,
        num=500,
        spider=False,
        interpol_kind='quadratic'
    ):
    """Smooth data

    x_list : list
        List with x values
    y_list : list
        List with y values
    num : int
        New number of interpolation points
    spider : bool
        Criteria whether spider plot or not
    interpol_kind : str
        Kind of interpolation, i.e. quadratic or cubic

    Note:
    ------
    - needs at least 4 entries in lists
    - https://docs.scipy.org/doc/scipy/reference/tutorial/interpolate.html
    - The smoothing prevents negative values by setting them to zero
    """
    if spider:

        min_x_val = min(x_list)
        max_x_val = math.pi * 2 #max is tow pi

        x_values = np.linspace(
            min_x_val,
            max_x_val,
            num=len(x_list),
            endpoint=True)

        f2 = interp1d(
            x_values,
            y_list,
            kind=interpol_kind) 

        x_smooth = np.linspace(
            min_x_val,
            max_x_val,
            num=num,
            endpoint=True)
    else:
        # Smooth x data
        x_smooth = np.linspace(
            min(x_list),
            max(x_list),
            num=num)

        f2 = interp1d(
            x_list,
            y_list,
            kind=interpol_kind)

    # smooth
    y_smooth = f2(x_smooth)

    # Prevent smoothing to go into negative values and replace negative values with zero

    # Get position of last negative entry
    try:
        pos_zero = int(np.argwhere(y_smooth < 0)[-1])
        y_smooth[:pos_zero] = 0
    except IndexError:
        pass

    return x_smooth, y_smooth

def smooth_line(
        input_x_line_data,
        input_y_line_data,
        nr_line_points=1000
    ):
    """https://stackoverflow.com/questions/5283649/plot-smooth-line-with-pyplot

    nr_line_points : int
        represents number of points to make between input_line_data.min and T.max
    """
    input_x_line_data = np.array(input_x_line_data)
    input_y_line_data = np.array(input_y_line_data)

    smooth_x_line_data = np.linspace(
        input_x_line_data.min(),
        input_x_line_data.max(),
        nr_line_points) 

    smooth_y_line_data = spline(
        input_x_line_data,
        input_y_line_data,
        smooth_x_line_data)

    return smooth_x_line_data, smooth_y_line_data



"""Testing
"""
from energy_demand.basic import lookup_tables
from energy_demand.assumptions import general_assumptions
from energy_demand.read_write import data_loader
'''
def test_load_non_param_assump():
    """
    """
    config_file = "C:/Users/cenv0553/ED/energy_demand/local_run_config_file.ini"
    config = data_loader.read_config_file(config_file)

    # Load data
    data = {}
    paths = config['CONFIG_DATA']
    lu = lookup_tables.basic_lookups()

    data['local_paths'] = config['DATA_PATHS']

    enduses, sectors, _, lookup_enduses, lookup_sector_enduses = data_loader.load_fuels(paths)

    general_assumptions.Assumptions(
        lookup_enduses=lookup_enduses,
        lookup_sector_enduses=lookup_sector_enduses,
        base_yr=2015,
        curr_yr=None,
        paths=paths,
        enduses=enduses,
        sectors=sectors)

def test_load_param_assump():
    """
    """
    config_file = "C:/Users/cenv0553/ED/energy_demand/local_run_config_file.ini"
    config = data_loader.read_config_file(config_file)

    # Load data
    config_data = config['CONFIG_DATA']
    data_loader.load_fuels(config_data)

    sim_param_expected = {}
    sim_param_expected['base_yr'] = 2015

    # Dummy test
    assert sim_param_expected['base_yr'] == 2015
'''